# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import sys
import time
from openc3.api.tlm_api import tlm
from openc3.script.exceptions import CheckError
from openc3.utilities.extract import (
    extract_fields_from_check_text,
    extract_fields_from_tlm_text,
)
from openc3.environment import OPENC3_SCOPE

DEFAULT_TLM_POLLING_RATE = 0.25

# NOTE: The formatting applied throughout uses :.Xf meaning X decimal points
# This allows extremely small wait times to simply be displayed 0.000.
# Without the 'f' :.X means display X significant figures


def check(*args, type="CONVERTED", scope="DEFAULT"):
    """Check the converted value of a telmetry item against a condition
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check(target_name, packet_name, item_name, comparison_to_eval)
    or
    check('target_name packet_name item_name > 1')
    """
    return _check(*args, type=type, scope=scope)


def check_raw(*args, scope="DEFAULT"):
    """Check the raw value of a telmetry item against a condition
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check(target_name, packet_name, item_name, comparison_to_eval)
    or
    check('target_name packet_name item_name > 1')
    """
    return _check(*args, type="RAW", scope=scope)


def check_formatted(*args, scope="DEFAULT"):
    """Check the formatted value of a telmetry item against a condition
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check(target_name, packet_name, item_name, comparison_to_eval)
    or
    check('target_name packet_name item_name > 1')
    """
    return _check(*args, type="FORMATTED", scope=scope)


def check_with_units(*args, scope="DEFAULT"):
    """Check the formatted with units value of a telmetry item against a condition
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check(target_name, packet_name, item_name, comparison_to_eval)
    or
    check('target_name packet_name item_name > 1')
    """
    return _check(*args, type="WITH_UNITS", scope=scope)


def check_exception(method_name, *args, **kwargs):
    """Executes the passed method and expects an exception to be raised.
    Raises a CheckError if an Exception is not raised.
    Usage: check_exception(method_name, method_params}"""
    try:
        method = method_name
        orig_kwargs = kwargs.copy()
        if "scope" not in kwargs:
            kwargs["scope"] = OPENC3_SCOPE
        getattr(sys.modules[__name__], method_name)(*args, **kwargs)
        method = f"{method_name}({', '.join(args)}"
        if orig_kwargs:
            method += f", {orig_kwargs}"
        method += ")"
    except Exception as error:
        print(f"CHECK: {method} raised {repr(error)}")
    else:
        raise CheckError(f"{method} should have raised an exception but did not.")


def check_tolerance(*args, type="CONVERTED", scope=OPENC3_SCOPE):
    """Check the converted value of a telmetry item against an expected value with a tolerance
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check_tolerance(target_name, packet_name, item_name, expected_value, tolerance)
    or
    check_tolerance('target_name packet_name item_name', expected_value, tolerance)
    """
    if type not in ["RAW", "CONVERTED"]:
        raise RuntimeError(f"Invalid type '{type}' for check_tolerance")

    (
        target_name,
        packet_name,
        item_name,
        expected_value,
        tolerance,
    ) = _check_tolerance_process_args(args)

    value = tlm(target_name, packet_name, item_name, type=type, scope=scope)
    if isinstance(value, list):
        expected_value, tolerance = _array_tolerance_process_args(
            len(value), expected_value, tolerance, "check_tolerance"
        )

        message = ""
        all_checks_ok = True
        for i in range(len(value)):
            range_bottom = expected_value[i] - tolerance[i]
            range_top = expected_value[i] + tolerance[i]
            check_str = f"CHECK: {_upcase(target_name, packet_name, item_name)}[{i}]"
            range_str = f"range {_frange(range_bottom)} to {_frange(range_top)} with value == {value[i]}"
            if value[i] >= range_bottom and value[i] <= range_top:
                message += f"{check_str} was within {range_str}\n"
            else:
                message += f"{check_str} failed to be within {range_str}\n"
                all_checks_ok = False

        if all_checks_ok:
            print(message)
        else:
            raise CheckError(message)
    else:
        range_bottom = expected_value - tolerance
        range_top = expected_value + tolerance
        check_str = f"CHECK: {_upcase(target_name, packet_name, item_name)}"
        range_str = f"range {_frange(range_bottom)} to {_frange(range_top)} with value == {value}"
        if value >= range_bottom and value <= range_top:
            print(f"{check_str} was within {range_str}")
        else:
            message = f"{check_str} failed to be within {range_str}"
            raise CheckError(message)


def check_expression(exp_to_eval, globals=None, locals=None):
    """Check to see if an expression is true without waiting.  If the expression
    is not true, the script will pause."""
    success = _openc3_script_wait_expression(exp_to_eval, 0, DEFAULT_TLM_POLLING_RATE, globals, locals)
    if success:
        print(f"CHECK: {exp_to_eval} is TRUE")
    else:
        message = f"CHECK: {exp_to_eval} is FALSE"
        raise CheckError(message)


def wait(*args, type="CONVERTED", quiet=False, scope=OPENC3_SCOPE):
    """Wait on an expression to be true.  On a timeout, the script will continue.
    Supports multiple signatures:
    wait(time)
    wait('target_name packet_name item_name > 1', timeout, polling_rate)
    wait('target_name', 'packet_name', 'item_name', comparison_to_eval, timeout, polling_rate)
    """
    time_diff = None

    match len(args):
        # wait() # indefinitely until they click Go
        case 0:
            start_time = time.time()
            openc3_script_sleep()
            time_diff = time.time() - start_time
            if not quiet:
                print(f"WAIT: Indefinite for actual time of {time_diff:.3f} seconds")
            return time_diff

        # wait(5) # absolute wait time
        case 1:
            try:
                value = float(args[0])
            except ValueError:
                raise RuntimeError("Non-numeric wait time specified")

            start_time = time.time()
            openc3_script_sleep(value)
            time_diff = time.time() - start_time
            if not quiet:
                print(f"WAIT: {value} seconds with actual time of {time_diff:.3f} seconds")
            return time_diff

        # wait('target_name packet_name item_name > 1', timeout, polling_rate) # polling_rate is optional
        case 2 | 3:
            (
                target_name,
                packet_name,
                item_name,
                comparison_to_eval,
            ) = extract_fields_from_check_text(args[0])
            timeout = args[1]
            if len(args) == 3:
                polling_rate = args[2]
            else:
                polling_rate = DEFAULT_TLM_POLLING_RATE
            return _execute_wait(
                target_name,
                packet_name,
                item_name,
                type,
                comparison_to_eval,
                timeout,
                polling_rate,
                quiet,
                scope,
            )

        # wait('target_name', 'packet_name', 'item_name', comparison_to_eval, timeout, polling_rate) # polling_rate is optional
        case 5 | 6:
            target_name = args[0]
            packet_name = args[1]
            item_name = args[2]
            comparison_to_eval = args[3]
            timeout = args[4]
            if len(args) == 6:
                polling_rate = args[5]
            else:
                polling_rate = DEFAULT_TLM_POLLING_RATE
            return _execute_wait(
                target_name,
                packet_name,
                item_name,
                type,
                comparison_to_eval,
                timeout,
                polling_rate,
                quiet,
                scope,
            )
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to wait()")


def wait_tolerance(*args, type="CONVERTED", quiet=False, scope=OPENC3_SCOPE):
    """Wait on an expression to be true.  On a timeout, the script will continue.
    Supports multiple signatures:
    wait_tolerance('target_name packet_name item_name', expected_value, tolerance, timeout, polling_rate)
    wait_tolerance('target_name', 'packet_name', 'item_name', expected_value, tolerance, timeout, polling_rate)
    """
    if type not in ["RAW", "CONVERTED"]:
        raise RuntimeError(f"Invalid type '{type}' for wait_tolerance")

    (
        target_name,
        packet_name,
        item_name,
        expected_value,
        tolerance,
        timeout,
        polling_rate,
    ) = _wait_tolerance_process_args(args, "wait_tolerance")
    start_time = time.time()
    value = tlm(target_name, packet_name, item_name, type=type, scope=scope)
    if isinstance(value, list):
        expected_value, tolerance = _array_tolerance_process_args(
            len(value), expected_value, tolerance, "wait_tolerance"
        )
        success, value = _openc3_script_wait_array_tolerance(
            len(value),
            target_name,
            packet_name,
            item_name,
            type,
            expected_value,
            tolerance,
            timeout,
            polling_rate,
        )
        time_diff = time.time() - start_time

        message = ""
        for i in range(0, len(value)):
            range_bottom = expected_value[i] - tolerance[i]
            range_top = expected_value[i] + tolerance[i]
            check_str = f"WAIT: {_upcase(target_name, packet_name, item_name)}[{i}]"
            range_str = f"range {_frange(range_bottom)} to {_frange(range_top)} with value == {value[i]} after waiting {time_diff:.3f} seconds"
            if value[i] >= range_bottom and value[i] <= range_top:
                message += f"{check_str} was within {range_str}\n"
            else:
                message += f"{check_str} failed to be within {range_str}\n"

        if not quiet:
            if success:
                print(message)
            else:
                print(f"WARN: {message}")
    else:
        success, value = _openc3_script_wait_tolerance(
            target_name,
            packet_name,
            item_name,
            type,
            expected_value,
            tolerance,
            timeout,
            polling_rate,
        )
        time_diff = time.time() - start_time
        range_bottom = expected_value - tolerance
        range_top = expected_value + tolerance
        wait_str = f"WAIT: {_upcase(target_name, packet_name, item_name)}"
        range_str = f"range {_frange(range_bottom)} to {_frange(range_top)} with value == {value} after waiting {time_diff:.3f} seconds"
        if not quiet:
            if success:
                print(f"{wait_str} was within {range_str}")
            else:
                print(f"WARN: {wait_str} failed to be within {range_str}")
    return success


def wait_expression(
    exp_to_eval,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
    globals=None,
    locals=None,
    quiet=False,
):
    """Wait on a custom expression to be true"""
    start_time = time.time()
    success = _openc3_script_wait_expression(exp_to_eval, timeout, polling_rate, globals, locals)
    time_diff = time.time() - start_time
    if not quiet:
        if success:
            print(f"WAIT: {exp_to_eval} is TRUE after waiting {time_diff:.3f} seconds")
        else:
            print(f"WARN: WAIT: {exp_to_eval} is FALSE after waiting {time_diff:.3f} seconds")
    return success


def wait_check(*args, type="CONVERTED", scope=OPENC3_SCOPE):
    """Wait for the converted value of a telmetry item against a condition or for a timeout
    and then check against the condition
    Supports two signatures:
    wait_check(target_name, packet_name, item_name, comparison_to_eval, timeout, polling_rate)
    or
    wait_check('target_name packet_name item_name > 1', timeout, polling_rate)"""
    (
        target_name,
        packet_name,
        item_name,
        comparison_to_eval,
        timeout,
        polling_rate,
    ) = _wait_check_process_args(args)
    start_time = time.time()
    success, value = _openc3_script_wait_value(
        target_name,
        packet_name,
        item_name,
        type,
        comparison_to_eval,
        timeout,
        polling_rate,
    )
    if isinstance(value, str):
        value = f"'{value}'"  # Show user the check against a quoted string
    time_diff = time.time() - start_time
    check_str = f"CHECK: {_upcase(target_name, packet_name, item_name)}"
    if comparison_to_eval:
        check_str += f" {comparison_to_eval}"
    with_value_str = f"with value == {value} after waiting {time_diff:.3f} seconds"
    if success:
        print(f"{check_str} success {with_value_str}")
    else:
        message = f"{check_str} failed {with_value_str}"
        raise CheckError(message)
    return time_diff


def wait_check_tolerance(*args, type="CONVERTED", scope=OPENC3_SCOPE):
    """Wait for the value of a telmetry item to be within a tolerance of a value
    and then check against the condition.
    Supports two signatures:
    wait_check_tolerance('target_name packet_name item_name', expected_value, tolerance, timeout, polling_rate)
    or
    wait_check_tolerance('target_name', 'packet_name', 'item_name', expected_value, tolerance, timeout, polling_rate)
    """
    if type not in ["RAW", "CONVERTED"]:
        raise RuntimeError(f"Invalid type '{type}' for wait_check_tolerance")

    (
        target_name,
        packet_name,
        item_name,
        expected_value,
        tolerance,
        timeout,
        polling_rate,
    ) = _wait_tolerance_process_args(args, "wait_check_tolerance")
    start_time = time.time()
    value = tlm(target_name, packet_name, item_name, type=type, scope=scope)
    if isinstance(value, list):
        expected_value, tolerance = _array_tolerance_process_args(
            len(value), expected_value, tolerance, "wait_check_tolerance"
        )
        success, value = _openc3_script_wait_array_tolerance(
            len(value),
            target_name,
            packet_name,
            item_name,
            type,
            expected_value,
            tolerance,
            timeout,
            polling_rate,
        )
        time_diff = time.time() - start_time

        message = ""
        for i in range(0, len(value)):
            range_bottom = expected_value[i] - tolerance[i]
            range_top = expected_value[i] + tolerance[i]
            check_str = f"CHECK: {_upcase(target_name, packet_name, item_name)}[{i}]"
            range_str = f"range {_frange(range_bottom)} to {_frange(range_top)} with value == {value[i]} after waiting {time_diff:.3f} seconds"
            if value[i] >= range_bottom and value[i] <= range_top:
                message += f"{check_str} was within {range_str}\n"
            else:
                message += f"{check_str} failed to be within {range_str}\n"

        if success:
            print(message)
        else:
            raise CheckError(message)
    else:
        success, value = _openc3_script_wait_tolerance(
            target_name,
            packet_name,
            item_name,
            type,
            expected_value,
            tolerance,
            timeout,
            polling_rate,
            scope,
        )
        time_diff = time.time() - start_time
        range_bottom = expected_value - tolerance
        range_top = expected_value + tolerance
        check_str = f"CHECK: {_upcase(target_name, packet_name, item_name)}"
        range_str = f"range {_frange(range_bottom)} to {_frange(range_top)} with value == {value} after waiting {time_diff:.3f} seconds"
        if success:
            print(f"{check_str} was within {range_str}")
        else:
            message = f"{check_str} failed to be within {range_str}"
            raise CheckError(message)
    return time_diff


def wait_check_expression(exp_to_eval, timeout, polling_rate=DEFAULT_TLM_POLLING_RATE, globals=None, locals=None):
    """Wait on an expression to be true.  On a timeout, the script will pause"""
    start_time = time.time()
    success = _openc3_script_wait_expression(exp_to_eval, timeout, polling_rate, globals, locals)
    time_diff = time.time() - start_time
    if success:
        print(f"CHECK: {exp_to_eval} is TRUE after waiting {time_diff:.3f} seconds")
    else:
        message = f"CHECK: {exp_to_eval} is FALSE after waiting {time_diff:.3f} seconds"
        raise CheckError(message)
    return time_diff


def wait_packet(
    target_name,
    packet_name,
    num_packets,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
    quiet=False,
    scope=OPENC3_SCOPE,
):
    success, _ = _wait_packet(
        False,
        target_name,
        packet_name,
        num_packets,
        timeout,
        polling_rate,
        quiet,
        scope,
    )
    return success


def wait_check_packet(
    target_name,
    packet_name,
    num_packets,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
    quiet=False,
    scope=OPENC3_SCOPE,
):
    """Wait for a telemetry packet to be received a certain number of times or timeout and raise an error"""
    _, time_diff = _wait_packet(True, target_name, packet_name, num_packets, timeout, polling_rate, quiet, scope)
    return time_diff


###########################################################################
# Private implementation details
###########################################################################


def openc3_script_sleep(sleep_time=None):
    if sleep_time:
        time.sleep(float(sleep_time))
    else:
        input("Press any key to continue...")


def _upcase(target_name, packet_name, item_name):
    """Creates a string with the parameters upcased"""
    return f"{target_name.upper()} {packet_name.upper()} {item_name.upper()}"


def _check(*args, type="CONVERTED", scope=OPENC3_SCOPE):
    """Implementation of the various check commands. It yields back to the
    caller to allow the return of the value through various telemetry calls.
    This method should not be called directly by application code."""
    target_name, packet_name, item_name, comparison_to_eval = _check_process_args(args, "check")
    value = tlm(target_name, packet_name, item_name, type=type, scope=scope)
    if comparison_to_eval:
        return _check_eval(target_name, packet_name, item_name, comparison_to_eval, value)
    else:
        print(f"CHECK: {_upcase(target_name, packet_name, item_name)} == {value}")


def _check_process_args(args, method_name):
    match len(args):
        case 1:
            (
                target_name,
                packet_name,
                item_name,
                comparison_to_eval,
            ) = extract_fields_from_check_text(args[0])
        case 3:
            target_name = args[0]
            packet_name = args[1]
            item_name = args[2]
            comparison_to_eval = None
        case 4:
            target_name = args[0]
            packet_name = args[1]
            item_name = args[2]
            comparison_to_eval = args[3]
        case _:
            # Invalid number of arguments
            raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to {method_name}()")
    if comparison_to_eval and not comparison_to_eval.isascii():
        raise RuntimeError(f"ERROR: Invalid comparison to non-ascii value: {comparison_to_eval}")
    return target_name, packet_name, item_name, comparison_to_eval


def _check_tolerance_process_args(args):
    length = len(args)
    if length == 3:
        target_name, packet_name, item_name = extract_fields_from_tlm_text(args[0])
        expected_value = args[1]
        if isinstance(args[2], list):
            tolerance = [abs(x) for x in args[2]]
        else:
            tolerance = abs(args[2])
    elif length == 5:
        target_name = args[0]
        packet_name = args[1]
        item_name = args[2]
        expected_value = args[3]
        if isinstance(args[4], list):
            tolerance = [abs(x) for x in args[4]]
        else:
            tolerance = abs(args[4])
    else:
        # Invalid number of arguments
        raise RuntimeError(f"ERROR: Invalid number of arguments ({length}) passed to check_tolerance()")
    return target_name, packet_name, item_name, expected_value, tolerance


def _wait_packet(
    check,
    target_name,
    packet_name,
    num_packets,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
    quiet=False,
    scope=OPENC3_SCOPE,
):
    """Wait for a telemetry packet to be received a certain number of times or timeout"""
    if check:
        type = "CHECK"
    else:
        type = "WAIT"
    initial_count = tlm(target_name, packet_name, "RECEIVED_COUNT", scope=scope)
    # If the packet has not been received the initial_count could be None
    if initial_count is None:
        initial_count = 0
    start_time = time.time()
    success, value = _openc3_script_wait_value(
        target_name,
        packet_name,
        "RECEIVED_COUNT",
        "CONVERTED",
        f">= {initial_count + num_packets}",
        timeout,
        polling_rate,
        scope,
    )
    # If the packet has not been received the value could be None
    if not value:
        value = 0
    time_diff = time.time() - start_time
    if success:
        if not quiet:
            print(
                f"{type}: {target_name.upper()} {packet_name.upper()} received {value - initial_count} times after waiting {time_diff:.3f} seconds"
            )
    else:
        message = f"{type}: {target_name.upper()} {packet_name.upper()} expected to be received {num_packets} times but only received {value - initial_count} times after waiting {time_diff:.3f} seconds"
        if check:
            raise CheckError(message)
        elif not quiet:
            print(f"WARN: {message}")
    return success, time_diff


def _execute_wait(
    target_name,
    packet_name,
    item_name,
    value_type,
    comparison_to_eval,
    timeout,
    polling_rate,
    quiet,
    scope,
):
    start_time = time.time()
    success, value = _openc3_script_wait_value(
        target_name,
        packet_name,
        item_name,
        value_type,
        comparison_to_eval,
        timeout,
        polling_rate,
        scope,
    )
    if isinstance(value, str):
        value = f"'{value}'"  # Show user the check against a quoted string
    time_diff = time.time() - start_time
    wait_str = f"WAIT: {_upcase(target_name, packet_name, item_name)} {comparison_to_eval}"
    value_str = f"with value == {value} after waiting {time_diff:.3f} seconds"
    if not quiet:
        if success:
            print(f"{wait_str} success {value_str}")
        else:
            print(f"WARN: {wait_str} failed {value_str}")
    return success


def _wait_tolerance_process_args(args, function_name):
    length = len(args)
    if length == 4 or length == 5:
        target_name, packet_name, item_name = extract_fields_from_tlm_text(args[0])
        expected_value = args[1]
        if isinstance(args[2], list):
            tolerance = [abs(x) for x in args[2]]
        else:
            tolerance = abs(args[2])
        timeout = args[3]
        if length == 5:
            polling_rate = args[4]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
    elif length == 6 or length == 7:
        target_name = args[0]
        packet_name = args[1]
        item_name = args[2]
        expected_value = args[3]
        if isinstance(args[4], list):
            tolerance = [abs(x) for x in args[4]]
        else:
            tolerance = abs(args[4])
        timeout = args[5]
        if length == 7:
            polling_rate = args[6]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
    else:
        # Invalid number of arguments
        raise RuntimeError(f"ERROR: Invalid number of arguments ({length}) passed to {function_name}()")
    return (
        target_name,
        packet_name,
        item_name,
        expected_value,
        tolerance,
        timeout,
        polling_rate,
    )


def _array_tolerance_process_args(array_size, expected_value, tolerance, function_name):
    """
    When testing an array with a tolerance, the expected value and tolerance
    can both be supplied as either an array or a single value.  If a single
    value is passed in, that value will be used for all array elements.
    """
    if isinstance(expected_value, list):
        if array_size != len(expected_value):
            raise RuntimeError(f"ERROR: Invalid array size for expected_value passed to {function_name}()")
    else:
        expected_value = [expected_value] * array_size
    if isinstance(tolerance, list):
        if array_size != len(tolerance):
            raise RuntimeError(f"ERROR: Invalid array size for tolerance passed to {function_name}()")
    else:
        tolerance = [tolerance] * array_size
    return expected_value, tolerance


def _wait_check_process_args(args):
    length = len(args)
    if length == 2 or length == 3:
        (
            target_name,
            packet_name,
            item_name,
            comparison_to_eval,
        ) = extract_fields_from_check_text(args[0])
        timeout = args[1]
        if length == 3:
            polling_rate = args[2]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
    elif length == 5 or length == 6:
        target_name = args[0]
        packet_name = args[1]
        item_name = args[2]
        comparison_to_eval = args[3]
        timeout = args[4]
        if length == 6:
            polling_rate = args[5]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
    else:
        # Invalid number of arguments
        raise RuntimeError(f"ERROR: Invalid number of arguments ({len(args)}) passed to wait_check()")
    return (
        target_name,
        packet_name,
        item_name,
        comparison_to_eval,
        timeout,
        polling_rate,
    )


def _openc3_script_wait(
    target_name,
    packet_name,
    item_name,
    value_type,
    timeout,
    polling_rate,
    exp_to_eval,
    scope,
):
    value = None
    end_time = time.time() + timeout
    if exp_to_eval and not exp_to_eval.isascii():
        raise RuntimeError("ERROR: Invalid comparison to non-ascii value")

    try:
        while True:
            work_start = time.time()
            value = tlm(target_name, packet_name, item_name, type=value_type, scope=scope)
            try:
                if eval(exp_to_eval):
                    return True, value
            # We get TypeError when trying to eval None >= 0 (for example)
            # In this case we just continue and see if eventually we get a good value from tlm()
            except TypeError:
                pass
            if time.time() >= end_time:
                break

            delta = time.time() - work_start
            sleep_time = polling_rate - delta
            end_delta = end_time - time.time()
            if end_delta < sleep_time:
                sleep_time = end_delta
            if sleep_time < 0:
                sleep_time = 0
            canceled = openc3_script_sleep(sleep_time)

            if canceled:
                value = tlm(target_name, packet_name, item_name, type=value_type, scope=scope)
                try:
                    if eval(exp_to_eval):
                        return True, value
                    else:
                        return False, value
                # We get TypeError when trying to eval None >= 0 (for example)
                except TypeError:
                    return False, value

    except NameError as error:
        parts = error.args[0].split("'")
        new_error = NameError(f"Uninitialized constant {parts[1]}. Did you mean '{parts[1]}' as a string?")
        raise new_error from error

    return False, value


# Wait for a converted telemetry item to pass a comparison
def _openc3_script_wait_value(
    target_name,
    packet_name,
    item_name,
    value_type,
    comparison_to_eval,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
    scope=OPENC3_SCOPE,
):
    if comparison_to_eval:
        exp_to_eval = "value " + comparison_to_eval
    else:
        exp_to_eval = None
    return _openc3_script_wait(
        target_name,
        packet_name,
        item_name,
        value_type,
        timeout,
        polling_rate,
        exp_to_eval,
        scope,
    )


def _openc3_script_wait_tolerance(
    target_name,
    packet_name,
    item_name,
    value_type,
    expected_value,
    tolerance,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
    scope=OPENC3_SCOPE,
):
    exp_to_eval = f"(value >= ({expected_value} - {abs(tolerance)}) and value <= ({expected_value} + {abs(tolerance)}))"
    return _openc3_script_wait(
        target_name,
        packet_name,
        item_name,
        value_type,
        timeout,
        polling_rate,
        exp_to_eval,
        scope,
    )


def _openc3_script_wait_array_tolerance(
    array_size,
    target_name,
    packet_name,
    item_name,
    value_type,
    expected_value,
    tolerance,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
    scope=OPENC3_SCOPE,
):
    statements = []
    for i in range(array_size):
        statements.append(
            f"(value[{i}] >= ({expected_value[i]} - {abs(tolerance[i])}) and value[{i}] <= ({expected_value[i]} + {abs(tolerance[i])}))"
        )
    exp_to_eval = " and ".join(statements)
    return _openc3_script_wait(
        target_name,
        packet_name,
        item_name,
        value_type,
        timeout,
        polling_rate,
        exp_to_eval,
        scope,
    )


def _openc3_script_wait_expression(exp_to_eval, timeout, polling_rate, globals=None, locals=None):
    """Wait on an expression to be true."""
    end_time = time.time() + timeout
    if not exp_to_eval.isascii():
        raise RuntimeError(f"Invalid comparison to non-ascii value: {exp_to_eval}")

    try:
        while True:
            work_start = time.time()
            if eval(exp_to_eval, globals, locals):
                return True
            if time.time() >= end_time:
                break

            delta = time.time() - work_start
            sleep_time = polling_rate - delta
            end_delta = end_time - time.time()
            if end_delta < sleep_time:
                sleep_time = end_delta
            if sleep_time < 0:
                sleep_time = 0
            canceled = openc3_script_sleep(sleep_time)

            if canceled:
                if eval(exp_to_eval, globals, locals):
                    return True
                else:
                    return None
    except NameError as error:
        parts = error.args[0].split("'")
        new_error = NameError(f"Uninitialized constant {parts[1]}. Did you mean '{parts[1]}' as a string?")
        raise new_error from error

    return None


def _check_eval(target_name, packet_name, item_name, comparison_to_eval, value):
    string = "value " + comparison_to_eval
    check_str = f"CHECK: {_upcase(target_name, packet_name, item_name)} {comparison_to_eval}"
    # Show user the check against a quoted string
    # Note: We have to preserve the original 'value' variable because we're going to eval against it
    if isinstance(value, str):
        value_str = f"'{value}'"
    else:
        value_str = value
    with_value = f"with value == {value_str}"
    try:
        if eval(string):
            print(f"{check_str} success {with_value}")
        else:
            message = f"{check_str} failed {with_value}"
            raise CheckError(message)
    except NameError as error:
        parts = error.args[0].split("'")
        new_error = NameError(f"Uninitialized constant {parts[1]}. Did you mean '{parts[1]}' as a string?")
        raise new_error from error


def _frange(value):
    if isinstance(value, float):
        # Display at most 6 significant figures on a range value
        # This truncates float values like 1.6500000000000001 to simply 1.65
        return f"{value:.6}"
    else:
        return value


# Interesting formatter to a specific number of significant digits:
# https://stackoverflow.com/questions/3410976/how-to-round-a-number-to-significant-figures-in-python?rq=3
# def format(value, sigfigs=9):
#     if isinstance(value, float):
#         return "{:.{p}g}".format(float("{:.{p}g}".format(value, p=sigfigs)), p=sigfigs)
#     else:
#         return value



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from datetime import datetime, timezone
from openc3.api import WHITELIST
from openc3.api.tlm_api import _tlm_process_args
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.topics.limits_event_topic import LimitsEventTopic
from openc3.models.cvt_model import CvtModel
from openc3.models.target_model import TargetModel

# from openc3.utilities.extract import *
from openc3.utilities.logger import Logger
from openc3.utilities.time import to_nsec_from_epoch

WHITELIST.extend(
    [
        "get_out_of_limits",
        "get_overall_limits_state",
        "limits_enabled",
        "enable_limits",
        "disable_limits",
        "get_limits",
        "set_limits",
        "get_limits_groups",
        "enable_limits_group",
        "disable_limits_group",
        "get_limits_sets",
        "set_limits_set",
        "get_limits_set",
        "get_limits_events",
    ]
)


# Return an array of arrays indicating all items in the packet that are out of limits
#   [[target name, packet name, item name, item limits state], ...]
#
# @return [Array<Array<String, String, String, String>>]
def get_out_of_limits(scope=OPENC3_SCOPE):
    authorize(permission="tlm", scope=scope)
    return LimitsEventTopic.out_of_limits(scope=scope)


# Get the overall limits state which is the worse case of all limits items.
# For example if any limits are YELLOW_LOW or YELLOW_HIGH then the overall limits state is YELLOW.
# If a single limit item then turns RED_HIGH the overall limits state is RED.
#
# @param ignored_items [Array<Array<String, String, String|nil>>] Array of [TGT, PKT, ITEM] strings
#   to ignore when determining overall state. Note, ITEM can be nil to indicate to ignore entire packet.
# @return [String] The overall limits state for the system, one of 'GREEN', 'YELLOW', 'RED'
def get_overall_limits_state(ignored_items=None, scope=OPENC3_SCOPE):
    # We only need to check out of limits items so call get_out_of_limits() which authorizes
    out_of_limits = get_out_of_limits(scope=scope)
    overall = "GREEN"

    # Build easily matchable ignore list
    if ignored_items is not None:
        new_items = []
        for item in ignored_items:
            if len(item) != 3:
                raise RuntimeError(f"Invalid ignored item: {item}. Must be [TGT, PKT, ITEM] where ITEM can be None.")
            if item[2] is None:
                item[2] = ""
            new_items.append("__".join(item))
        ignored_items = new_items
    else:
        ignored_items = []

    for target_name, packet_name, item_name, limits_state in out_of_limits:
        # Ignore this item if we match one of the ignored items
        for item in ignored_items:
            if item in f"{target_name}__{packet_name}__{item_name}":
                break
        else:  # Executed if 'for item in ignored_items:' did NOT break
            if limits_state == "RED" or limits_state == "RED_HIGH" or limits_state == "RED_LOW":
                overall = limits_state
                break  # Red is as high as we go so no need to look for more

            # If our overall state is currently blue or green we can go to any state
            if overall in ["BLUE", "GREEN", "GREEN_HIGH", "GREEN_LOW"]:
                overall = limits_state
            # else YELLOW - Stay at YELLOW until we find a red

    if overall == "GREEN_HIGH" or overall == "GREEN_LOW" or overall == "BLUE":
        overall = "GREEN"
    if overall == "YELLOW_HIGH" or overall == "YELLOW_LOW":
        overall = "YELLOW"
    if overall == "RED_HIGH" or overall == "RED_LOW":
        overall = "RED"
    return overall


# Whether the limits are enabled for the given item
#
# Accepts two different calling styles:
#   limits_enabled("TGT PKT ITEM")
#   limits_enabled('TGT','PKT','ITEM')
#
# Favor the first syntax where possible as it is more succinct.
#
# @param args [String|Array<String>] See the description for calling style
# @return [Boolean] Whether limits are enable for the item
def limits_enabled(*args, scope=OPENC3_SCOPE):
    target_name, packet_name, item_name = _tlm_process_args(args, "limits_enabled", scope=scope)
    authorize(permission="tlm", target_name=target_name, packet_name=packet_name, scope=scope)
    item = TargetModel.packet_item(target_name, packet_name, item_name, scope=scope)
    if item["limits"].get("enabled"):
        return True
    else:
        return False


# Enable limits checking for a telemetry item
#
# Accepts two different calling styles:
#   enable_limits("TGT PKT ITEM")
#   enable_limits('TGT','PKT','ITEM')
#
# Favor the first syntax where possible as it is more succinct.
#
# @param args [String|Array<String>] See the description for calling style
def enable_limits(*args, scope=OPENC3_SCOPE):
    target_name, packet_name, item_name = _tlm_process_args(args, "enable_limits", scope=scope)
    authorize(
        permission="tlm_set",
        target_name=target_name,
        packet_name=packet_name,
        scope=scope,
    )
    packet = TargetModel.packet(target_name, packet_name, scope=scope)
    found_item = None
    for item in packet["items"]:
        if item["name"] == item_name:
            item["limits"]["enabled"] = True
            found_item = item
            break
    if found_item is None:
        raise RuntimeError(f"Item '{target_name} {packet_name} {item_name}' does not exist")

    TargetModel.set_packet(target_name, packet_name, packet, scope=scope)

    message = f"Enabling Limits For '{target_name} {packet_name} {item_name}'"
    Logger.info(message, scope=scope)

    event = {
        "type": "LIMITS_ENABLE_STATE",
        "target_name": target_name,
        "packet_name": packet_name,
        "item_name": item_name,
        "enabled": True,
        "time_nsec": to_nsec_from_epoch(datetime.now(timezone.utc)),
        "message": message,
    }
    LimitsEventTopic.write(event, scope=scope)


# Disable limit checking for a telemetry item
#
# Accepts two different calling styles:
#   disable_limits("TGT PKT ITEM")
#   disable_limits('TGT','PKT','ITEM')
#
# Favor the first syntax where possible as it is more succinct.
#
# @param args [String|Array<String>] See the description for calling style
def disable_limits(*args, scope=OPENC3_SCOPE):
    target_name, packet_name, item_name = _tlm_process_args(args, "disable_limits", scope=scope)
    authorize(
        permission="tlm_set",
        target_name=target_name,
        packet_name=packet_name,
        scope=scope,
    )
    packet = TargetModel.packet(target_name, packet_name, scope=scope)
    found_item = None
    for item in packet["items"]:
        if item["name"] == item_name:
            item["limits"].pop("enabled", None)
            found_item = item
            break
    if found_item is None:
        raise RuntimeError(f"Item '{target_name} {packet_name} {item_name}' does not exist")

    TargetModel.set_packet(target_name, packet_name, packet, scope=scope)

    message = f"Disabling Limits for '{target_name} {packet_name} {item_name}'"
    Logger.info(message, scope=scope)

    event = {
        "type": "LIMITS_ENABLE_STATE",
        "target_name": target_name,
        "packet_name": packet_name,
        "item_name": item_name,
        "enabled": False,
        "time_nsec": to_nsec_from_epoch(datetime.now(timezone.utc)),
        "message": message,
    }
    LimitsEventTopic.write(event, scope=scope)


# Get a Hash of all the limits sets defined for an item. Hash keys are the limit
# set name in uppercase (note there is always a DEFAULT) and the value is an array
# of limit values: red low, yellow low, yellow high, red high, <green low, green high>.
# Green low and green high are optional.
#
# For example: {'DEFAULT' => [-80, -70, 60, 80, -20, 20],
#               'TVAC' => [-25, -10, 50, 55] }
#
# @return [Hash{String => Array<Number, Number, Number, Number, Number, Number>}]
def get_limits(target_name, packet_name, item_name, scope=OPENC3_SCOPE):
    authorize(permission="tlm", target_name=target_name, packet_name=packet_name, scope=scope)
    limits = {}
    item = _get_item(target_name, packet_name, item_name, scope=scope)
    for key, vals in item["limits"].items():
        if not isinstance(vals, dict):
            continue

        limits[key] = [
            vals["red_low"],
            vals["yellow_low"],
            vals["yellow_high"],
            vals["red_high"],
        ]
        if vals.get("green_low"):
            limits[key] += [vals["green_low"], vals["green_high"]]
    return limits


# Change the limits settings for a given item. By default, a new limits set called 'CUSTOM'
# is created to avoid overriding existing limits.
def set_limits(
    target_name,
    packet_name,
    item_name,
    red_low,
    yellow_low,
    yellow_high,
    red_high,
    green_low=None,
    green_high=None,
    limits_set="CUSTOM",
    persistence=None,
    enabled=True,
    scope=OPENC3_SCOPE,
):
    authorize(
        permission="tlm_set",
        target_name=target_name,
        packet_name=packet_name,
        scope=scope,
    )
    if (red_low > yellow_low) or (yellow_low >= yellow_high) or (yellow_high > red_high):
        raise RuntimeError("Invalid limits specified. Ensure yellow limits are within red limits.")
    if (green_low and green_high) and (
        (yellow_low > green_low) or (green_low >= green_high) or (green_high > yellow_high)
    ):
        raise RuntimeError("Invalid limits specified. Ensure green limits are within yellow limits.")
    packet = TargetModel.packet(target_name, packet_name, scope=scope)
    found_item = None
    for item in packet["items"]:
        if item["name"] == item_name:
            if item["limits"]:
                if persistence:
                    item["limits"]["persistence_setting"] = persistence
                if enabled:
                    item["limits"]["enabled"] = True
                else:
                    item["limits"].pop("enabled", None)
                limits = {}
                limits["red_low"] = red_low
                limits["yellow_low"] = yellow_low
                limits["yellow_high"] = yellow_high
                limits["red_high"] = red_high
                if green_low and green_high:
                    limits["green_low"] = green_low
                if green_low and green_high:
                    limits["green_high"] = green_high
                item["limits"][limits_set] = limits
                found_item = item
                break
            else:
                raise RuntimeError("Cannot set_limits on item without any limits")
    if found_item is None:
        raise RuntimeError(f"Item '{target_name} {packet_name} {item_name}' does not exist")
    message = (
        f"Setting '{target_name} {packet_name} {item_name}' limits to {red_low} {yellow_low} {yellow_high} {red_high}"
    )
    if green_low and green_high:
        message += f" {green_low} {green_high}"
    message += f" in set {limits_set} with persistence {persistence} as enabled {enabled}"
    Logger.info(message, scope=scope)

    TargetModel.set_packet(target_name, packet_name, packet, scope=scope)

    event = {
        "type": "LIMITS_SETTINGS",
        "target_name": target_name,
        "packet_name": packet_name,
        "item_name": item_name,
        "red_low": red_low,
        "yellow_low": yellow_low,
        "yellow_high": yellow_high,
        "red_high": red_high,
        "green_low": green_low,
        "green_high": green_high,
        "limits_set": limits_set,
        "persistence": persistence,
        "enabled": enabled,
        "time_nsec": to_nsec_from_epoch(datetime.now(timezone.utc)),
        "message": message,
    }
    LimitsEventTopic.write(event, scope=scope)


# Returns all limits_groups and their members
# @since 5.0.0 Returns hash with values
# @return [Hash{String => Array<Array<String, String, String>>]
def get_limits_groups(scope=OPENC3_SCOPE):
    authorize(permission="tlm", scope=scope)
    return TargetModel.limits_groups(scope=scope)


# Enables limits for all the items in the group
#
# @param group_name [String] Name of the group to enable
def enable_limits_group(group_name, scope=OPENC3_SCOPE):
    _limits_group(group_name, action="enable", scope=scope)


# Disables limits for all the items in the group
#
# @param group_name [String] Name of the group to disable
def disable_limits_group(group_name, scope=OPENC3_SCOPE):
    _limits_group(group_name, action="disable", scope=scope)


# Returns all defined limits sets
#
# @return [Array<String>] All defined limits sets
def get_limits_sets(scope=OPENC3_SCOPE):
    authorize(permission="tlm", scope=scope)
    sets = list(LimitsEventTopic.sets(scope=scope).keys())
    sets.sort()
    return sets


# Changes the active limits set that applies to all telemetry
#
# @param limits_set [String] The name of the limits set
def set_limits_set(limits_set, scope=OPENC3_SCOPE):
    authorize(permission="tlm_set", scope=scope)
    message = f"Setting Limits Set: {limits_set}"
    Logger.info(message, scope=scope)
    LimitsEventTopic.write(
        {
            "type": "LIMITS_SET",
            "set": str(limits_set),
            "time_nsec": to_nsec_from_epoch(datetime.now(timezone.utc)),
            "message": message,
        },
        scope=scope,
    )


# Returns the active limits set that applies to all telemetry
#
# @return [String] The current limits set
def get_limits_set(scope=OPENC3_SCOPE):
    authorize(permission="tlm", scope=scope)
    return LimitsEventTopic.current_set(scope=scope)


# Returns limits events starting at the provided offset. Passing nil for an
# offset will return the last received limits event and associated offset.
#
# @param offset [Integer] Offset to start reading limits events. Nil to return
#   the last received limits event (if any).
# @param count [Integer] The total number of events returned. Default is 100.
# @return [Hash, Integer] Event hash followed by the offset. The offset can
#   be used in subsequent calls to return events from where the last call left off.
def get_limits_events(offset=None, count=100, scope=OPENC3_SCOPE):
    authorize(permission="tlm", scope=scope)
    return LimitsEventTopic.read(offset, count=count, scope=scope)


# Enables or disables a limits group
def _limits_group(group_name, action, scope):
    authorize(permission="tlm_set", scope=scope)
    group_name.upper()
    group = get_limits_groups(scope=scope).get(group_name)
    if group is None:
        raise RuntimeError(
            f"LIMITS_GROUP {group_name} undefined. Ensure your telemetry definition contains the line: LIMITS_GROUP {group_name}"
        )

    Logger.info(f"{action.capitalize()} Limits Group: {group_name}", scope=scope)
    last_target_name = None
    last_packet_name = None
    packet = None
    for target_name, packet_name, item_name in group:
        if last_target_name != target_name or last_packet_name != packet_name:
            if last_target_name and last_packet_name:
                TargetModel.set_packet(last_target_name, last_packet_name, packet, scope=scope)
            packet = TargetModel.packet(target_name, packet_name, scope=scope)
        for item in packet["items"]:
            if item["name"] == item_name:
                if action == "enable":
                    enabled = True
                    item["limits"]["enabled"] = True
                    message = f"Enabling Limits for '{target_name} {packet_name} {item_name}'"
                elif action == "disable":
                    enabled = False
                    item["limits"].pop("enabled", None)
                    message = f"Disabling Limits for '{target_name} {packet_name} {item_name}'"
                Logger.info(message, scope=scope)

                event = {
                    "type": "LIMITS_ENABLE_STATE",
                    "target_name": target_name,
                    "packet_name": packet_name,
                    "item_name": item_name,
                    "enabled": enabled,
                    "time_nsec": to_nsec_from_epoch(datetime.now(timezone.utc)),
                    "message": message,
                }
                LimitsEventTopic.write(event, scope=scope)
                break
        last_target_name = target_name
        last_packet_name = packet_name
    if last_target_name and last_packet_name:
        TargetModel.set_packet(last_target_name, last_packet_name, packet, scope=scope)


# Gets an item. The code below is mostly duplicated from tlm_process_args in tlm_api.rb.
#
# @param target_name [String] target name
# @param packet_name [String] packet name
# @param item_name [String] item name
# @param scope [String] scope
# @return Hash The requested item based on the packet name
def _get_item(target_name, packet_name, item_name, cache_timeout=0.1, scope=OPENC3_SCOPE):
    # Determine if this item exists, it will raise appropriate errors if not
    if packet_name == "LATEST":
        packet_name = CvtModel.determine_latest_packet_for_item(target_name, item_name, cache_timeout, scope)
    return TargetModel.packet_item(target_name, packet_name, item_name, scope=scope)



# Copyright 2023, OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.io.io_multiplexer import IoMultiplexer


# Adds STDOUT to the multiplexed streams
class Stdout(IoMultiplexer):
    my_instance = None

    def __init__(self):
        super().__init__()
        self.streams.append(self.STDOUT)
        Stdout.my_instance = self

    # @return [Stdout] Returns a single instance of Stdout
    @classmethod
    def instance(cls):
        if not Stdout.my_instance:
            cls()
        return Stdout.my_instance

    @property
    def encoding(self):
        return self.STDOUT.encoding



# Copyright 2025 OpenC3, Inc
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
from openc3.api import WHITELIST
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.models.offline_access_model import OfflineAccessModel
from openc3.utilities.authentication import (
    OpenC3KeycloakAuthentication,
)

try:
    from openc3enterprise.utilities.authorization import user_info
except ImportError:
    from openc3.utilities.authorization import user_info

WHITELIST.extend(["offline_access_needed", "set_offline_access"])


def offline_access_needed(manual=False, scope=OPENC3_SCOPE, token=None):
    authorize(permission="system", manual=manual, scope=scope)
    try:
        authorize(permission="script_view", manual=manual, scope=scope)
    except Exception:
        # Not needed if can't run scripts
        return False
    info = user_info(token)
    if "offline_access" in info["roles"]:
        username = info["username"]
        if username and username != "":
            model = OfflineAccessModel.get_model(name=username, scope=scope)
            if model and model.offline_access_token:
                auth = OpenC3KeycloakAuthentication(os.environ.get("OPENC3_KEYCLOAK_URL"))
                valid_token = auth.get_token_from_refresh_token(model.offline_access_token)
                if valid_token:
                    return False
                else:
                    model.offline_access_token = None
                    model.update()
                    return True
            return True
        else:
            return False
    else:
        return False


def set_offline_access(offline_access_token, manual=False, scope=OPENC3_SCOPE, token=None):
    authorize(permission="script_view", manual=manual, scope=scope)
    info = user_info(token)
    username = info["username"]
    if not username or username == "":
        raise ValueError("Invalid username")
    model = OfflineAccessModel.get_model(name=username, scope=scope)
    if model:
        model.offline_access_token = offline_access_token
        model.update()
    else:
        model = OfflineAccessModel(name=username, offline_access_token=offline_access_token, scope=scope)
        model.create()



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.processors.processor import Processor


class WatermarkProcessor(Processor):
    # @param item_name [String] The name of the item to gather statistics on
    # @param value_type #See Processor::initialize
    def __init__(self, item_name, value_type="CONVERTED"):
        super().__init__(value_type)
        self.item_name = str(item_name).upper()
        self.reset()

    # Run watermarks on the item
    #
    # See Processor#call
    def call(self, packet, buffer):
        value = packet.read(self.item_name, self.value_type, buffer)
        high_water = self.results.get("HIGH_WATER")
        if high_water is None or value > high_water:
            self.results["HIGH_WATER"] = value
        low_water = self.results.get("LOW_WATER")
        if low_water is None or value < low_water:
            self.results["LOW_WATER"] = value

    # Reset any state
    def reset(self):
        self.results["HIGH_WATER"] = None
        self.results["LOW_WATER"] = None

    # Convert to configuration file string
    def to_config(self):
        return f"  PROCESSOR {self.name} {self.__class__.__name__} {self.item_name} {self.value_type}\n"

    def as_json(self):
        return {
            "name": self.name,
            "class": self.__class__.__name__.name,
            "params": [self.item_name, self.value_type],
        }



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import queue
import os
import pathlib
import gzip
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.polling import PollingObserver
from openc3.top_level import get_class_from_module
from openc3.interfaces.interface import Interface
from openc3.config.config_parser import ConfigParser
from openc3.utilities.string import class_name_to_filename
from openc3.utilities.string import build_timestamped_filename
from openc3.utilities.sleeper import Sleeper


class NewFileEventHandler(FileSystemEventHandler):
    def __init__(self, interface):
        self.interface = interface

    def on_any_event(self, event: FileSystemEvent) -> None:
        if event.event_type == "created" or (
            event.event_type == "moved" and event.dest_path == self.interface.telemetry_read_folder
        ):
            self.interface.queue.put(event.src_path)


class FileInterface(Interface):
    # @param command_write_folder [String] Folder to write command files to - Set to None to disallow writes
    # @param telemetry_read_folder [String] Folder to read telemetry files from - Set to None to disallow reads
    # @param telemetry_archive_folder [String] Folder to move read telemetry files to - Set to DELETE to delete files
    # @param file_read_size [Integer] Number of bytes to read from the file at a time
    # @param stored [Boolean] Whether to set stored flag on read telemetry
    # @param protocol_type [String] Name of the protocol to use
    #   with this interface
    # @param protocol_args [Array<String>] Arguments to pass to the protocol
    def __init__(
        self,
        command_write_folder,
        telemetry_read_folder,
        telemetry_archive_folder,
        file_read_size=65536,
        stored=True,
        protocol_type=None,
        protocol_args=[],
    ):
        super().__init__()

        self.protocol_type = ConfigParser.handle_none(protocol_type)
        self.protocol_args = protocol_args
        if self.protocol_type:
            protocol_class_name = str(protocol_type).capitalize() + "Protocol"
            filename = class_name_to_filename(protocol_class_name)
            klass = get_class_from_module(f"openc3.interfaces.protocols.{filename}", protocol_class_name)
            self.add_protocol(klass, protocol_args, "PARAMS")

        self.command_write_folder = ConfigParser.handle_none(command_write_folder)
        self.telemetry_read_folder = ConfigParser.handle_none(telemetry_read_folder)
        self.telemetry_archive_folder = ConfigParser.handle_none(telemetry_archive_folder)
        self.file_read_size = int(file_read_size)
        self.stored = ConfigParser.handle_true_false(stored)

        if not self.telemetry_read_folder:
            self.read_allowed = False
        if not self.command_write_folder:
            self.write_allowed = False
        if not self.command_write_folder:
            self.write_raw_allowed = False

        self.file = None
        self.file_path = None
        self.listener = None
        self._connected = False
        self.extension = ".bin"
        self.label = "command"
        self.queue = queue.Queue()
        self.polling = False
        self.recursive = False
        self.throttle = None
        self.discard_file_header_bytes = None
        self.sleeper = None

    def connect(self):
        super().connect() # Reset the protocols

        if self.telemetry_read_folder:
            event_handler = NewFileEventHandler(self)
            if self.polling:
                self.listener = PollingObserver()
            else:
                self.listener = Observer()
            self.listener.schedule(event_handler, self.telemetry_read_folder, recursive=self.recursive)
            self.listener.start()

        self._connected = True

    def connected(self):
        return self._connected

    def disconnect(self):
        if self.file and not self.file.closed:
            self.file.close()
        self.file = None
        self.file_path = None
        if self.listener:
            self.listener.stop()
        if self.sleeper:
            self.sleeper.cancel()
        self.listener = None
        self.queue.put(None)
        super().disconnect()
        self._connected = False

    def read_interface(self):
        while True:
            if self.file:
                # Read more data from existing file
                data = self.file.read(self.file_read_size)
                if self.throttle and self.sleeper.sleep(self.throttle):
                    return None, None
                if data is not None and len(data) > 0:
                    self.read_interface_base(data, None)
                    return data, None
                else:
                    self.finish_file()

            # Find the next file to read
            file = self.get_next_telemetry_file()
            if file:
                if file.endswith(".gz"):
                    self.file = gzip.open(file, 'rb')
                else:
                    self.file = open(file, 'rb')
                self.file_path = file
                if self.discard_file_header_bytes is not None:
                    self.file.read(self.discard_file_header_bytes)
                continue

            # Wait for a file to read
            result = self.queue.get()
            if result is None:
                return None, None

    def write_interface(self, data, extra=None):
        # Write this data into its own file
        with open(self.create_unique_filename(), "wb") as file:
            file.write(data)

        self.write_interface_base(data, extra)
        return data, extra

    def convert_data_to_packet(self, data, extra=None):
        packet = super().convert_data_to_packet(data, extra)
        if packet and self.stored:
            packet.stored = True

        return packet

    # Supported Options
    # LABEL - Label to add to written files
    # EXTENSION - Extension to add to written files
    # (see Interface#set_option)
    def set_option(self, option_name, option_values):
        super().set_option(option_name, option_values)
        match option_name.upper():
            case "LABEL":
                self.label = option_values[0]
            case "EXTENSION":
                self.extension = option_values[0]
            case "POLLING":
                self.polling = ConfigParser.handle_true_false(option_values[0])
            case "RECURSIVE":
                self.recursive = ConfigParser.handle_true_false(option_values[0])
            case "THROTTLE":
                self.throttle = float(option_values[0])
                self.sleeper = Sleeper()
            case "DISCARD_FILE_HEADER_BYTES":
                self.discard_file_header_bytes = int(option_values[0])

    def finish_file(self):
        self.file.close()
        self.file = None

        # Archive (or DELETE) complete file
        if self.telemetry_archive_folder == "DELETE":
            os.remove(self.file_path)
        else:
            new_path = os.path.join(self.telemetry_archive_folder, os.path.basename(self.file_path))
            os.rename(self.file_path, new_path)
        self.file_path = None

    def get_next_telemetry_file(self):
        if self.recursive:
            path = pathlib.Path(self.telemetry_read_folder)
            list = [str(item) for item in path.rglob("*") if item.is_file()]
        else:
            list = os.listdir(self.telemetry_read_folder)
        list.sort()
        if len(list) > 0:
            if self.recursive:
                return list[0]
            else:
                return os.path.join(self.telemetry_read_folder, list[0])
        else:
            return None

    def create_unique_filename(self):
        # Create a filename that doesn't exist
        attempt = None
        while True:
            filename = os.path.join(
                self.command_write_folder,
                build_timestamped_filename([self.label, attempt], self.extension),
            )
            if os.path.isfile(filename):
                if attempt is None:
                    attempt = 0
                attempt += 1
            else:
                return filename



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.system.system import System
from openc3.config.config_parser import ConfigParser
from openc3.utilities.logger import Logger
from openc3.interfaces.protocols.protocol import Protocol
from queue import SimpleQueue, Empty
import time


# Protocol that waits for a response for any commands with a defined response packet.
# The response packet is identified but not defined by the protocol.
class CmdResponseProtocol(Protocol):
    # @param response_timeout [Float] Number of seconds to wait before timing out
    #   when waiting for a response
    # @param response_polling_period [Float] Number of seconds to wait between polling
    #   for a response
    # @param raise_exceptions [String] Whether to raise exceptions when errors
    #   occur in the protocol like unexpected responses or response timeouts.
    # @param allow_empty_data [true/false/nil] See Protocol#initialize
    def __init__(
        self, response_timeout=5.0, response_polling_period=0.02, raise_exceptions=False, allow_empty_data=None
    ):
        super().__init__(allow_empty_data)
        self.response_timeout = ConfigParser.handle_none(response_timeout)
        if self.response_timeout:
            self.response_timeout = float(response_timeout)
        self.response_polling_period = float(response_polling_period)
        self.raise_exceptions = ConfigParser.handle_true_false(raise_exceptions)
        self.write_block_queue = SimpleQueue()
        self.response_packet = None

    def connect_reset(self):
        super().connect_reset()
        try:
            while self.write_block_queue.qsize != 0:
                self.write_block_queue.get_nowait()
        except Empty:
            pass

    def disconnect_reset(self):
        super().disconnect_reset()
        self.write_block_queue.put(None)  # Unblock the write block queue

    def read_packet(self, packet):
        if self.response_packet is not None:
            # Grab the response packet specified in the command
            result_packet = System.telemetry.packet(self.response_packet[0], self.response_packet[1]).clone()
            result_packet.buffer = packet.buffer
            result_packet.received_time = None
            result_packet.stored = packet.stored
            result_packet.extra = packet.extra

            # Release the write
            self.write_block_queue.put(None)

            # This returns the fully identified and defined packet
            # Received time is handled by the interface microservice
            return result_packet
        else:
            return packet

    def write_packet(self, packet):
        # Setup the response packet (if there is one)
        # This primes waiting for the response in post_write_interface
        self.response_packet = packet.response

        return packet

    def post_write_interface(self, packet, data, extra=None):
        if self.response_packet is not None:
            if self.response_timeout:
                response_timeout_time = time.time() + self.response_timeout
            else:
                response_timeout_time = None

            # Block the write until the response is received
            while True:
                try:
                    self.write_block_queue.get_nowait()
                    break
                except Empty:
                    time.sleep(self.response_polling_period)
                    if response_timeout_time is None:
                        continue
                    if response_timeout_time and time.time() < response_timeout_time:
                        continue
                    interface_name = ""
                    if self.interface is not None:
                        interface_name = self.interface.name
                    self.handle_error(f"{interface_name}: Timeout waiting for response")

            self.response_packet = None
        return super().post_write_interface(packet, data, extra)

    def handle_error(self, msg):
        Logger.error(msg)
        if self.raise_exceptions:
            raise RuntimeError(msg)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import re
import copy
from openc3.packets.structure_item import StructureItem
from openc3.packets.packet_item_limits import PacketItemLimits
from openc3.utilities.string import quote_if_necessary, simple_formatted
from openc3.conversions.conversion import Conversion


class PacketItem(StructureItem):
    # The allowable state colors
    STATE_COLORS = ["GREEN", "YELLOW", "RED"]

    def __init__(
        self,
        name,
        bit_offset,
        bit_size,
        data_type,
        endianness,
        array_size=None,
        overflow="ERROR",
    ):
        super().__init__(name, bit_offset, bit_size, data_type, endianness, array_size, overflow)
        self.format_string = None
        self.read_conversion = None
        self.write_conversion = None
        self.id_value = None
        self.states = None
        self.description = None
        self.units_full = None
        self.units = None
        self.default = None
        # self.range = None
        self.minimum = None
        self.maximum = None
        self.required = False
        self.hazardous = None
        self.messages_disabled = None
        self.state_colors = None
        self.limits = PacketItemLimits()
        self.persistence_setting = 1
        self.persistence_count = 0
        self.meta = None

    @property
    def format_string(self):
        return self.__format_string

    @format_string.setter
    def format_string(self, format_string):
        if format_string:
            if not isinstance(format_string, str):
                raise TypeError(
                    f"{self.name}: format_string must be a str but is a {format_string.__class__.__name__}"
                )
            if not re.search(r"%.*(b|B|d|i|o|u|x|X|e|E|f|g|G|a|A|c|p|s|%)", format_string):
                raise ValueError(f"{self.name}: format_string invalid '{format_string}'")
            self.__format_string = format_string
        else:
            self.__format_string = None

    @property
    def read_conversion(self):
        return self.__read_conversion

    @read_conversion.setter
    def read_conversion(self, read_conversion):
        if read_conversion:
            if not isinstance(read_conversion, Conversion):
                raise TypeError(
                    f"{self.name}: read_conversion must be a Conversion but is a {read_conversion.__class__.__name__}"
                )
            self.__read_conversion = read_conversion
        else:
            self.__read_conversion = None

    @property
    def write_conversion(self):
        return self.__write_conversion

    @write_conversion.setter
    def write_conversion(self, write_conversion):
        if write_conversion:
            if not isinstance(write_conversion, Conversion):
                raise TypeError(
                    f"{self.name}: write_conversion must be a Conversion but is a {write_conversion.__class__.__name__}"
                )
            self.__write_conversion = write_conversion
        else:
            self.__write_conversion = None

    @property
    def id_value(self):
        return self.__id_value

    @id_value.setter
    def id_value(self, id_value):
        if id_value is not None:
            self.__id_value = self.convert(id_value, self.data_type)
        else:
            self.__id_value = None

    # Assignment operator for states to make sure it is a dict with uppercase keys
    @property
    def states(self):
        return self.__states

    @states.setter
    def states(self, states):
        if states is not None:
            if not isinstance(states, dict):
                raise TypeError(f"{self.name}: states must be a dict but is a {states.__class__.__name__}")

            # Make sure all states are in upper case
            self.__states = {}
            self.__states_by_value = {}
            for key, value in states.items():
                upper = key.upper()
                self.__states[upper] = value
                self.__states_by_value[value] = upper
            if self.state_colors is None:
                self.state_colors = {}
        else:
            self.__states = None
            self.__states_by_value = None

    def states_by_value(self):
        return self.__states_by_value

    @property
    def description(self):
        return self.__description

    @description.setter
    def description(self, description):
        if description:
            if not isinstance(description, str):
                raise TypeError(
                    f"{self.name}: description must be a str but is a {description.__class__.__name__}"
                )
            self.__description = description
        else:
            self.__description = None

    @property
    def units_full(self):
        return self.__units_full

    @units_full.setter
    def units_full(self, units_full):
        if units_full:
            if not isinstance(units_full, str):
                raise TypeError(f"{self.name}: units_full must be a str but is a {units_full.__class__.__name__}")
            self.__units_full = units_full
        else:
            self.__units_full = None

    @property
    def units(self):
        return self.__units

    @units.setter
    def units(self, units):
        if units:
            if not isinstance(units, str):
                raise TypeError(f"{self.name}: units must be a str but is a {units.__class__.__name__}")
            self.__units = units
        else:
            self.__units = None

    def check_default_and_range_data_types(self):
        if self.default and not self.write_conversion:
            if self.array_size is not None:
                if not isinstance(self.default, list):
                    raise TypeError(
                        f"{self.name}: default must be a list but is a {self.default.__class__.__name__}"
                    )
            else:
                match self.data_type:
                    case "INT" | "UINT":
                        if not isinstance(self.default, int):
                            raise TypeError(
                                f"{self.name}: default must be a int but is a {self.default.__class__.__name__}"
                            )
                        if not isinstance(self.minimum, int):
                            raise TypeError(
                                f"{self.name}: minimum must be a int but is a {self.minimum.__class__.__name__}"
                            )
                        if not isinstance(self.maximum, int):
                            raise TypeError(
                                f"{self.name}: maximum must be a int but is a {self.maximum.__class__.__name__}"
                            )
                    case "FLOAT":
                        if not isinstance(self.default, (float, int)):
                            raise TypeError(
                                f"{self.name}: default must be a float but is a {self.default.__class__.__name__}"
                            )

                        self.default = float(self.default)

                        if not isinstance(self.minimum, (int, float)):
                            raise TypeError(
                                f"{self.name}: minimum must be a float but is a {self.minimum.__class__.__name__}"
                            )
                        if not isinstance(self.maximum, (int, float)):
                            raise TypeError(
                                f"{self.name}: maximum must be a float but is a {self.maximum.__class__.__name__}"
                            )
                    case "BLOCK" | "STRING":
                        if not isinstance(self.default, (str, bytes, bytearray)):
                            raise TypeError(
                                f"{self.name}: default must be a str but is a {self.default.__class__.__name__}"
                            )

    @property
    def hazardous(self):
        return self.__hazardous

    @hazardous.setter
    def hazardous(self, hazardous):
        if hazardous is not None:
            if not isinstance(hazardous, dict):
                raise TypeError(f"{self.name}: hazardous must be a dict but is a {hazardous.__class__.__name__}")
            self.__hazardous = hazardous
        else:
            self.__hazardous = None

    @property
    def messages_disabled(self):
        return self.__messages_disabled

    @messages_disabled.setter
    def messages_disabled(self, messages_disabled):
        if messages_disabled is not None:
            if not isinstance(messages_disabled, dict):
                raise TypeError(
                    f"{self.name}: messages_disabled must be a dict but is a {messages_disabled.__class__.__name__}"
                )

            self.__messages_disabled = messages_disabled
        else:
            self.__messages_disabled = None

    @property
    def state_colors(self):
        return self.__state_colors

    @state_colors.setter
    def state_colors(self, state_colors):
        if state_colors is not None:
            if not isinstance(state_colors, dict):
                raise TypeError(
                    f"{self.name}: state_colors must be a dict but is a {state_colors.__class__.__name__}"
                )

            self.__state_colors = state_colors
        else:
            self.__state_colors = None

    @property
    def limits(self):
        return self.__limits

    @limits.setter
    def limits(self, limits):
        if limits is not None:
            if not isinstance(limits, PacketItemLimits):
                raise TypeError(
                    f"{self.name}: limits must be a PacketItemLimits but is a {limits.__class__.__name__}"
                )

            self.__limits = limits
        else:
            self.__limits = None

    @property
    def meta(self):
        return self.__meta

    @meta.setter
    def meta(self, meta):
        if meta is not None:
            if not isinstance(meta, dict):
                raise TypeError(f"{self.name}: meta must be a dict but is a {meta.__class__.__name__}")

            self.__meta = meta
        else:
            self.__meta = {}

    # Make a light weight clone of this item
    def clone(self):
        item = copy.copy(self)
        # Since we're copying and not calling the constructor
        # we have to manually update the create_index
        item.create_index = StructureItem.create_index
        StructureItem.create_index += 1
        return item

    # def clone(self):
    #   item = super().clone()
    #   item.format_string = self.format_string.clone if self.format_string:
    #   item.read_conversion = self.read_conversion.clone if self.read_conversion:
    #   item.write_conversion = self.write_conversion.clone if self.write_conversion:
    #   item.states = self.states.clone if self.states:
    #   item.description = self.description.clone if self.description:
    #   item.units_full = self.units_full.clone if self.units_full:
    #   item.units = self.units.clone if self.units:
    #   item.default = self.default.clone if self.default and String === self.default:
    #   item.hazardous = self.hazardous.clone if self.hazardous:
    #   item.messages_disabled = self.messages_disabled.clone if self.messages_disabled:
    #   item.state_colors = self.state_colors.clone if self.state_colors:
    #   item.limits = self.limits.clone if self.limits:
    #   item.meta = self.meta.clone if self.meta:
    #   item

    def to_hash(self):
        item = {}
        item["format_string"] = self.format_string
        if self.read_conversion:
            item["read_conversion"] = str(self.read_conversion)
        else:
            item["read_conversion"] = None
        if self.write_conversion:
            item["write_conversion"] = str(self.write_conversion)
        else:
            item["write_conversion"] = None
        item["id_value"] = self.id_value
        item["states"] = self.states
        item["description"] = self.description
        item["units_full"] = self.units_full
        item["units"] = self.units
        item["default"] = self.default
        # item["range"] = self.range
        item["minimum"] = self.minimum
        item["maximum"] = self.maximum
        item["required"] = self.required
        item["hazardous"] = self.hazardous
        item["messages_disabled"] = self.messages_disabled
        item["state_colors"] = self.state_colors
        item["limits"] = self.limits.as_json()
        item["meta"] = None
        if self.meta:
            item["meta"] = self.meta
        return item

    # def calculate_range(self):
    #     first = self.range.first
    #     last = self.range.last
    #     if self.data_type == "FLOAT":
    #         if self.bit_size == 32:
    #             if self.range.first == -3.402823e38:
    #                 first = "MIN"
    #             if self.range.last == 3.402823e38:
    #                 last = "MAX"
    #         else:
    #             if self.range.first == -sys.float_info.max:
    #                 first = "MIN"
    #             if self.range.last == sys.float_info.max:
    #                 last = "MAX"
    #     return [first, last]

    def to_config(self, cmd_or_tlm, default_endianness):
        config = ""
        if self.description:
            description = self.description.replace('"', "'")
        else:
            description = ""
        if cmd_or_tlm == "TELEMETRY":
            if self.array_size is not None:
                config += f'  ARRAY_ITEM {quote_if_necessary(self.name)} {self.bit_offset} {self.bit_size} {self.data_type} {self.array_size} "{description}"'
            elif self.id_value:
                id_value = self.id_value
                if self.data_type == "BLOCK" or self.data_type == "STRING":
                    if not self.id_value.isascii():
                        id_value = "0x" + simple_formatted(self.id_value)
                    else:
                        id_value = f'"{self.id_value}"'
                config += f'  ID_ITEM {quote_if_necessary(self.name)} {self.bit_offset} {self.bit_size} {self.data_type} {id_value} "{description}"'
            else:
                config += f'  ITEM {quote_if_necessary(self.name)} {self.bit_offset} {self.bit_size} {self.data_type} "{description}"'
        else:  # 'COMMAND'
            if self.array_size is not None:
                config += f'  ARRAY_PARAMETER {quote_if_necessary(self.name)} {self.bit_offset} {self.bit_size} {self.data_type} {self.array_size} "{description}"'
            else:
                config += self.parameter_config()
        if self.endianness != default_endianness and self.data_type != "STRING" and self.data_type != "BLOCK":
            config += " {self.endianness}"
        config += "\n"

        if self.required:
            config += "    REQUIRED\n"
        if self.format_string:
            config += f"    FORMAT_STRING {quote_if_necessary(self.format_string)}\n"
        if self.units:
            config += f"    UNITS {quote_if_necessary(self.units_full)} {quote_if_necessary(self.units)}\n"
        if self.overflow != "ERROR":
            config += f"    OVERFLOW {self.overflow}\n"

        if self.states:
            for state_name, state_value in self.states.items():
                config += f"    STATE {quote_if_necessary(state_name)} {quote_if_necessary(str(state_value))}"
                if self.hazardous and self.hazardous[state_name]:
                    config += f" HAZARDOUS {quote_if_necessary(self.hazardous[state_name])}"
                if self.messages_disabled and self.messages_disabled[state_name]:
                    config += " DISABLE_MESSAGES"
                if self.state_colors and self.state_colors[state_name]:
                    config += f" {self.state_colors[state_name]}"
                config += "\n"

        if self.read_conversion:
            config += self.read_conversion.to_config("READ")
        if self.write_conversion:
            config += self.write_conversion.to_config("WRITE")

        if self.limits:
            if self.limits.values:
                for limits_set, limits_values in self.limits.values.items():
                    config += f"    LIMITS {limits_set} {self.limits.persistence_setting} {'ENABLED' if self.limits.enabled else 'DISABLED'} {limits_values[0]} {limits_values[1]} {limits_values[2]} {limits_values[3]}"
                    if len(limits_values) > 4:
                        config += f" {limits_values[4]} {limits_values[5]}\n"
                    else:
                        config += "\n"
            if self.limits.response:
                config += self.limits.response.to_config

        if self.meta:
            for key, values in self.meta.items():
                vals = " ".join([quote_if_necessary(x) for x in values])
                config += f"    META {quote_if_necessary(key)} {vals}\n"

        return config

    def as_json(self):
        config = {}
        config["name"] = self.name
        config["bit_offset"] = self.bit_offset
        config["bit_size"] = self.bit_size
        config["data_type"] = self.data_type
        if self.array_size is not None:
            config["array_size"] = self.array_size
        config["description"] = self.description
        if self.id_value is not None:
            config["id_value"] = self.id_value
        if self.default is not None:
            config["default"] = self.default
        config["minimum"] = self.minimum
        config["maximum"] = self.maximum
        config["endianness"] = self.endianness
        config["required"] = self.required
        if self.format_string:
            config["format_string"] = self.format_string
        if self.units:
            config["units"] = self.units
            config["units_full"] = self.units_full
        config["overflow"] = self.overflow
        if self.states:
            states = {}
            config["states"] = states
            for state_name, state_value in self.states.items():
                state = {}
                states[state_name] = state
                state["value"] = state_value
                if self.hazardous and self.hazardous.get(state_name) is not None:
                    state["hazardous"] = self.hazardous[state_name]
                if self.messages_disabled and self.messages_disabled.get(state_name) is not None:
                    state["messages_disabled"] = self.messages_disabled[state_name]
                if self.state_colors and self.state_colors.get(state_name) is not None:
                    state["color"] = self.state_colors[state_name]

        if self.read_conversion:
            config["read_conversion"] = self.read_conversion.as_json()
        if self.write_conversion:
            config["write_conversion"] = self.write_conversion.as_json()

        if self.limits:
            config["limits"] = {}
            if self.limits.enabled:
                config["limits"]["enabled"] = True
            else:
                config["limits"]["enabled"] = False
            if self.limits.values:
                config["limits"]["persistence_setting"] = self.limits.persistence_setting
                if self.limits.response:
                    config["limits"]["response"] = self.limits.response
                for limits_set, limits_values in self.limits.values.items():
                    limits = {}
                    limits["red_low"] = limits_values[0]
                    limits["yellow_low"] = limits_values[1]
                    limits["yellow_high"] = limits_values[2]
                    limits["red_high"] = limits_values[3]
                    if len(limits_values) > 4:
                        limits["green_low"] = limits_values[4]
                        limits["green_high"] = limits_values[5]
                    config["limits"][limits_set] = limits
            if self.limits.response:
                config["limits_response"] = self.limits.response.as_json()

        if self.meta:
            config["meta"] = self.meta
        return config

    @classmethod
    def from_json(cls, item_hash):
        # Convert strings to symbols
        endianness = item_hash.get("endianness")
        data_type = item_hash.get("data_type")
        overflow = item_hash.get("overflow")
        item = PacketItem(
            item_hash.get("name"),
            item_hash.get("bit_offset"),
            item_hash.get("bit_size"),
            data_type,
            endianness,
            item_hash.get("array_size"),
            overflow,
        )
        item.description = item_hash.get("description")
        item.id_value = item_hash.get("id_value")
        item.default = item_hash.get("default")
        item.minimum = item_hash.get("minimum")
        item.maximum = item_hash.get("maximum")
        item.required = item_hash.get("required")
        item.format_string = item_hash.get("format_string")
        item.units = item_hash.get("units")
        item.units_full = item_hash.get("units_full")
        if item_hash.get("states"):
            item.states = {}
            item.hazardous = {}
            item.messages_disabled = {}
            item.state_colors = {}
            for state_name, state in item_hash.get("states").items():
                item.states[state_name] = state.get("value")
                item.hazardous[state_name] = state.get("hazardous")
                item.messages_disabled[state_name] = state.get("messages_disabled")
                item.state_colors[state_name] = state.get("color")
        # Recreate OpenC3 built-in conversions
        #   if item_hash.get('read_conversion']:
        #     try:
        #       item.read_conversion = OpenC3::const_get(item_hash['read_conversion']['class'])(*item_hash['read_conversion']['params'])
        #     except:
        #       Logger.error(f"{item.name} read_conversion of {item_hash['read_conversion']} could not be instantiated due to {error}")
        #   if item_hash.get('write_conversion']:
        #     try:
        #       item.write_conversion = OpenC3::const_get(item_hash['write_conversion']['class'])(*item_hash['write_conversion']['params'])
        #     except:
        #       Logger.error(f"{item.name} write_conversion of {item_hash['write_conversion']} could not be instantiated due to {error}")

        item.limits = PacketItemLimits()
        if item_hash.get("limits"):
            # Delete the keys so the only ones left are limits sets
            persistence_setting = item_hash["limits"].pop("persistence_setting", None)
            if persistence_setting:
                item.limits.persistence_setting = persistence_setting
            item_hash["limits"].pop("response", None)  # Can't round trip response
            if item_hash["limits"].pop("enabled", None):
                item.limits.enabled = True
            values = {}
            for set, items in item_hash["limits"].items():
                values[set] = [
                    items["red_low"],
                    items["yellow_low"],
                    items["yellow_high"],
                    items["red_high"],
                ]
                if items.get("green_low") is not None and items.get("green_high") is not None:
                    values[set] += [items["green_low"], items["green_high"]]
            if len(values) > 0:
                item.limits.values = values
        item.meta = item_hash.get("meta")
        return item

    def parameter_config(self):
        if self.id_value:
            value = self.id_value
            config = "  ID_PARAMETER "
        else:
            value = self.default
            config = "  PARAMETER "
        config += f"{quote_if_necessary(self.name)} {self.bit_offset} {self.bit_size} {self.data_type} "

        if self.data_type == "BLOCK" or self.data_type == "STRING":
            if not value.isascii():
                val_string = "0x" + simple_formatted(value)
            else:
                val_string = f'"{value}"'
        else:
            # first, last = self.calculate_range()
            # config += f"{first} {last} "
            config += f"{self.minimum} {self.maximum} "
            val_string = value
        if self.description:
            description = self.description.replace('"', "'")
            config += f'{val_string} "{description}"'
        else:
            config += f"{val_string}"
        return config

    # Convert a value into the given data type
    def convert(self, value, data_type):
        try:
            match data_type:
                case "INT" | "UINT":
                    return int(value)
                case "FLOAT":
                    return float(value)
                case "STRING" | "BLOCK":
                    return value
        except ValueError:
            raise ValueError(f"{self.name}: Invalid value: {value} for data type: {data_type}")



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


from datetime import datetime, timezone

NSEC_PER_SECOND = 1_000_000_000


def from_nsec_from_epoch(nsec_from_epoch):
    if nsec_from_epoch is None:
        nsec_from_epoch = datetime().now(timezone.utc)
    return datetime.fromtimestamp(nsec_from_epoch / NSEC_PER_SECOND, timezone.utc)


def to_nsec_from_epoch(time):
    if time is None:
        return 0
    return int(time.timestamp()) * NSEC_PER_SECOND + time.microsecond * 1000


# @return [String] Date formatted as YYYYMMDDHHmmSSNNNNNNNNN
def to_timestamp(time):
    return time.strftime("%Y%m%d%H%M%S%f000")


def formatted(time):
    return time.strftime("%Y/%m/%d %H:%M:%S.%f")[:-3]



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


class AuthError(RuntimeError):
    pass


class ForbiddenError(RuntimeError):
    pass


# All the authorization is done by Ruby code in the Ruby API backend.
# This code is basically a NOOP for now. If we ever want to build a whole
# new API endpoint in Python we'll have to implement an Enterprise
# authorize() like in Ruby.
def authorize(
    permission=None,
    target_name=None,
    packet_name=None,
    interface_name=None,
    router_name=None,
    manual=False,
    scope=None,
):
    if not scope:
        raise AuthError("Scope is required")


def user_info(_token):
    return {}  # EE does stuff here



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import time
from threading import Lock

class ThreadManager:
    MONITOR_SLEEP_SECONDS = 0.25

    # Variable that holds the singleton instance
    instance_obj = None

    # Mutex used to ensure that only one instance is created
    instance_mutex = Lock()

    # Get the singleton instance of ThreadManager
    @classmethod
    def instance(cls):
        if ThreadManager.instance_obj is not None:
            return ThreadManager.instance_obj

        with ThreadManager.instance_mutex:
            ThreadManager.instance_obj = cls()
            return ThreadManager.instance_obj

    def __init__(self):
        self.threads = []
        self.shutdown_started = False

    def register(self, thread, stop_object=None, shutdown_object=None):
        self.threads.append([thread, stop_object, shutdown_object])

    def monitor(self):
        while True:
            for thread, _, _ in self.threads:
                if not thread.is_alive():
                    return
            time.sleep(self.MONITOR_SLEEP_SECONDS)

    def shutdown(self):
        with ThreadManager.instance_mutex:
            if self.shutdown_started:
                return
            self.shutdown_started = True
        for thread, stop_object, shutdown_object in self.threads:
            if thread.is_alive():
                if stop_object is not None:
                    stop_object.stop()
                if shutdown_object is not None:
                    shutdown_object.shutdown()

    def join(self):
        for thread, _, _ in self.threads:
            thread.join()



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


from openc3.conversions.unix_time_conversion import UnixTimeConversion
from openc3.utilities.time import formatted


# Converts a unix format time: Epoch Jan 1 1970, seconds and microseconds,
# into a formatted string.
class UnixTimeFormattedConversion(UnixTimeConversion):
    # Initializes converted_type to :STRING and converted_bit_size to 0
    #
    # @param seconds_item_name [String] The telemetry item in the packet which
    #   represents the number of seconds since the UNIX time epoch
    # @param microseconds_item_name [String] The telemetry item in the packet
    #   which represents microseconds
    def __init__(self, seconds_item_name, microseconds_item_name=None):
        # self.params is set by the parent class in super()
        super().__init__(seconds_item_name, microseconds_item_name)
        self.converted_type = "STRING"
        self.converted_bit_size = 0

    # @param (see Conversion#call)
    # @return [String] Formatted packet time
    def call(self, value, packet, buffer):
        return formatted(super().call(value, packet, buffer))

    # @return [String] The name of the class followed by the time conversion
    def __str__(self):
        result = f"UnixTimeFormattedConversion {self.seconds_item_name}"
        if self.microseconds_item_name:
            result += f" {self.microseconds_item_name}"
        return result



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.config.config_parser import ConfigParser
from openc3.conversions.conversion import Conversion
from openc3.system.system import System


class ObjectReadConversion(Conversion):
    def __init__(self, cmd_or_tlm, target_name, packet_name):
        super().__init__()
        cmd_or_tlm = ConfigParser.handle_none(cmd_or_tlm)
        if cmd_or_tlm:
            self.cmd_or_tlm = str(cmd_or_tlm).upper()
            if self.cmd_or_tlm not in ["CMD", "TLM", "COMMAND", "TELEMETRY"]:
                raise TypeError(f"Unknown type: {cmd_or_tlm}")
        else:
            # Unknown - Will need to search
            self.cmd_or_tlm = None
        self.target_name = str(target_name).upper()
        self.packet_name = str(packet_name).upper()
        self.converted_type = "OBJECT"
        self.converted_bit_size = 0
        self.params = [self.cmd_or_tlm, self.target_name, self.packet_name]

    def lookup_packet(self):
        if self.cmd_or_tlm:
            if self.cmd_or_tlm == "CMD" or self.cmd_or_tlm == "COMMAND":
                return System.commands.packet(self.target_name, self.packet_name)
            else:
                return System.telemetry.packet(self.target_name, self.packet_name)
        else:
            # Always searches commands first
            try:
                return System.commands.packet(self.target_name, self.packet_name)
            except RuntimeError:
                return System.telemetry.packet(self.target_name, self.packet_name)

    # Perform the conversion on the value.
    def call(self, value, _packet, buffer):
        fill_packet = self.lookup_packet()
        fill_packet.buffer = value
        all = fill_packet.read_all("CONVERTED", buffer, True)
        return {item[0]: item[1] for item in all}

    # @return [String] The conversion class
    def __str__(self):
        return f"{self.__class__.__name__} {self.cmd_or_tlm if self.cmd_or_tlm else 'None'} {self.target_name} {self.packet_name}"

    # @return [String] Config fragment for this conversion
    def to_config(self, read_or_write):
        return f"{read_or_write}_CONVERSION openc3/conversions/object_read_conversion.py {self.cmd_or_tlm if self.cmd_or_tlm else 'None'} {self.target_name} {self.packet_name}\n"



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import atexit
import tempfile
import threading
import traceback
import json
from openc3.system.system import System
from openc3.utilities.bucket import Bucket
from openc3.utilities.logger import Logger
from openc3.utilities.metric import Metric
from openc3.utilities.secrets import Secrets
from openc3.utilities.sleeper import Sleeper
from openc3.utilities.store import EphemeralStore
from openc3.utilities.thread_manager import ThreadManager
from openc3.topics.topic import Topic
from openc3.environment import OPENC3_CONFIG_BUCKET
from openc3.models.microservice_model import MicroserviceModel
from openc3.models.microservice_status_model import MicroserviceStatusModel

# TODO:
# OpenC3.require_file 'openc3/utilities/open_telemetry'

openc3_scope = "DEFAULT"


class Microservice:
    @classmethod
    def class_run_body(cls, microservice):
        try:
            MicroserviceStatusModel.set(microservice.as_json(), scope=microservice.scope)
            microservice.state = "RUNNING"
            microservice.run()
            microservice.state = "FINISHED"
        except Exception as err:
            # TODO: Handle SystemExit and SignalException
            # if SystemExit === err or SignalException === err:
            #   microservice.state = 'KILLED'
            # else:
            if microservice:
                microservice.error = err
                microservice.state = "DIED_ERROR"
            Logger.fatal(f"Microservice {microservice.name} dying from exception\n{traceback.format_exc()}")
            microservice.shutdown() # Dying in crash so should try to shutdown
        finally:
            if microservice:
                MicroserviceStatusModel.set(microservice.as_json(), scope=microservice.scope)

    @classmethod
    def class_run(cls, name=None):
        microservice = None
        if name is None:
            name = os.environ.get("OPENC3_MICROSERVICE_NAME")
        microservice = cls(name)
        thread = threading.Thread(target=cls.class_run_body, args=[microservice], daemon=True)
        thread.start()
        ThreadManager.instance().register(thread, shutdown_object=microservice)
        ThreadManager.instance().monitor()
        ThreadManager.instance().shutdown()

    def as_json(self):
        json = {
            "name": self.name,
            "state": self.state,
            "count": self.count,
            "plugin": self.plugin,
        }
        if self.error is not None:
            json["error"] = repr(self.error)
        if self.custom is not None:
            json["custom"] = self.custom.as_json()
        return json

    def __init__(self, name, is_plugin=False):
        self.shutdown_complete = False
        if name is None:
            raise RuntimeError("Microservice must be named")

        self.name = name
        split_name = name.split("__")
        if len(split_name) != 3:
            raise RuntimeError(f"Name {name} doesn't match convention of SCOPE__TYPE__NAME")

        self.scope = split_name[0]
        global openc3_scope
        openc3_scope = self.scope
        self.cancel_thread = False
        self.metric = Metric(microservice=self.name, scope=self.scope)
        Logger.scope = self.scope
        Logger.microservice_name = self.name
        self.logger = Logger()
        self.logger.scope = self.scope
        self.logger.microservice_name = self.name
        self.secrets = Secrets.getClient()

        # OpenC3.setup_open_telemetry(self.name, False)

        # Create temp folder for this microservice
        self.temp_dir = tempfile.TemporaryDirectory()

        # Get microservice configuration from Redis
        self.config = MicroserviceModel.get(self.name, scope=self.scope)
        if self.config:
            self.topics = self.config["topics"]
            self.plugin = self.config["plugin"]
            if self.config["secrets"]:
                self.secrets.setup(self.config["secrets"])
        else:
            self.config = {}
            self.plugin = None
        self.logger.info(f"Microservice initialized with config:\n{self.config}")
        if not hasattr(self, "topics") or self.topics is None:
            self.topics = []
        self.microservice_topic = f"MICROSERVICE__{self.name}"

        # Get configuration for any targets
        self.target_names = self.config.get("target_names")
        if self.target_names is None:
            self.target_names = []
        if not is_plugin:
            System.setup_targets(self.target_names, self.temp_dir.name, scope=self.scope)

        # Use atexit to shutdown cleanly no matter how we die
        atexit.register(self.shutdown)

        self.count = 0
        self.error = None
        self.custom = None
        self.state = "INITIALIZED"
        self.work_dir = self.config.get("work_dir")

        if is_plugin:
            self.config["cmd"]

            # Get Microservice files from bucket storage
            temp_dir = tempfile.TemporaryDirectory()
            bucket = OPENC3_CONFIG_BUCKET
            client = Bucket.getClient()

            prefix = f"{self.scope}/microservices/{self.name}/"
            file_count = 0
            for object in client.list_objects(bucket=bucket, prefix=prefix):
                response_target = os.path.join(temp_dir, object.key.split(prefix)[-1])
                os.makedirs(os.path.dirname(response_target), exist_ok=True)
                client.get_object(bucket=bucket, key=object.key, path=response_target)
                file_count += 1

            # Adjust @work_dir to microservice files downloaded if files and a relative path
            if file_count > 0 and self.work_dir[0] != "/":
                self.work_dir = os.path.join(temp_dir, self.work_dir)

            # TODO: Check Syntax on any python files
            # ruby_filename = None
            #  for part in cmd_array:
            #   if /\.rb$/.match?(part):
            #     ruby_filename = part
            #     break
            # if ruby_filename:
            #   OpenC3.set_working_dir(self.work_dir) do
            #     if os.path.exist(ruby_filename):
            #       # Run ruby syntax so we can log those
            #       syntax_check, _ = Open3.capture2e("ruby -c {ruby_filename}")
            #       if /Syntax OK/.match?(syntax_check):
            #         self.logger.info("Ruby microservice {self.name} file {ruby_filename} passed syntax check\n", scope: self.scope)
            #       else:
            #         self.logger.error("Ruby microservice {self.name} file {ruby_filename} failed syntax check\n{syntax_check}", scope: self.scope)
            #     else:
            #       self.logger.error("Ruby microservice {self.name} file {ruby_filename} does not exist", scope: self.scope)
        else:
            self.microservice_status_sleeper = Sleeper()
            self.microservice_status_period_seconds = 5
            self.microservice_status_thread = threading.Thread(target=self._status_thread, daemon=True)
            self.microservice_status_thread.start()
            ThreadManager.instance().register(self.microservice_status_thread)

    # Must be implemented by a subclass
    def run(self):
        self.shutdown()

    def shutdown(self, state='STOPPED'):
        if self.shutdown_complete:
            return  # Nothing more to do
        self.logger.info(f"Shutting down microservice: {self.name}")
        self.state = state
        self.cancel_thread = True
        if self.microservice_status_sleeper:
            self.microservice_status_sleeper.cancel()
        MicroserviceStatusModel.set(self.as_json(), scope=self.scope)
        if self.temp_dir is not None:
            self.temp_dir.cleanup()
        self.metric.shutdown()
        self.logger.debug(f"Shutting down microservice complete: {self.name}")
        self.shutdown_complete = True

    def setup_microservice_topic(self):
        self.topics.append(self.microservice_topic)
        thread_id = threading.get_native_id()
        ephemeral_store_instance = EphemeralStore.instance()
        if thread_id not in ephemeral_store_instance.topic_offsets:
            ephemeral_store_instance.topic_offsets[thread_id] = {}
        ephemeral_store_instance.topic_offsets[thread_id][self.microservice_topic] = "0-0"

    # Returns if the command was handled
    def microservice_cmd(self, topic, msg_id, msg_hash, _):
        command = msg_hash.get("command")
        if command == "ADD_TOPICS" and msg_hash.get("topics"):
            topics = json.loads(msg_hash["topics"])
            if topics and isinstance(topics, list):
                for new_topic in topics:
                    if new_topic not in self.topics:
                        self.topics.append(new_topic)
            else:
                raise RuntimeError(f"Invalid topics given to microservice_cmd: {topics}")
            Topic.trim_topic(topic, msg_id)
            return True
        Topic.trim_topic(topic, msg_id)
        return False

    def _status_thread(self):
        while not self.cancel_thread:
            try:
                MicroserviceStatusModel.set(self.as_json(), scope=self.scope)
                if self.microservice_status_sleeper.sleep(self.microservice_status_period_seconds):
                    break
            except RuntimeError as error:
                self.logger.error(f"{self.name} status thread died: {repr(error)}")
                raise error



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import json
from openc3.utilities.extract import *
import openc3.script
from openc3.environment import OPENC3_SCOPE


def get_screen_list(scope: str = OPENC3_SCOPE):
    """Gets a list of screens

    Args:
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        The json result of the method call
    """
    try:
        endpoint = "/openc3-api/screens"
        # Pass the name of the ENV variable name where we pull the actual bucket name
        response = openc3.script.API_SERVER.request("get", endpoint, scope=scope)
        if not response or response.status_code != 200:
            raise RuntimeError(f"Unexpected response to get_screen_list: {response}")
        screen_list = {}
        filenames = json.loads(response.text)
        for filename in filenames:
            # TARGET/screens/filename.txt
            split_filename = filename.split("/")
            target_name = split_filename[0]
            screen_name = os.path.splitext(os.path.basename(filename))[0].upper()
            if not screen_list.get(target_name):
                screen_list[target_name] = []
            screen_list[target_name].append(screen_name)
        return screen_list
    except Exception as error:
        raise RuntimeError(f"get_screen_list failed due to {repr(error)}") from error


def get_screen_definition(target_name: str, screen_name: str, scope: str = OPENC3_SCOPE):
    """Gets a screen definition

    Args:
        target_name (str) the target name
        screen_name (str) the screen name
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        The json result of the method call
    """
    try:
        endpoint = f"/openc3-api/screen/{target_name.upper()}/{screen_name.upper()}"
        response = openc3.script.API_SERVER.request(
            "get",
            endpoint,
            headers={
                "Accept": "text/plain",
            },
            scope=scope,
        )
        if not response or response.status_code != 200:
            raise RuntimeError(f"Screen definition not found: {target_name} {screen_name}")

        return response.text
    except Exception as error:
        raise RuntimeError(f"get_screen_definition failed due to {repr(error)}") from error


def create_screen(target_name: str, screen_name: str, definition: str, scope: str = OPENC3_SCOPE):
    """Create a screen definition

    Args:
        target_name (str) the target name
        screen_name (str) the screen name
        definition (str) the screen definition
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        The json result of the method call
    """
    try:
        endpoint = "/openc3-api/screen"
        data = {"target": target_name, "screen": screen_name, "text": definition}
        response = openc3.script.API_SERVER.request("post", endpoint, data=data, json=True, scope=scope)
        if not response or response.status_code != 200:
            if response:
                parsed = json.loads(response.text)
                raise RuntimeError(f"create_screen error: {parsed['error']}")
            else:
                raise RuntimeError("create_screen failed")
        return response.text
    except Exception as error:
        raise RuntimeError(f"create_screen failed due to {repr(error)}") from error


def delete_screen(target_name: str, screen_name: str, scope: str = OPENC3_SCOPE):
    """Create a screen definition

    Args:
        target_name (str) the target name
        screen_name (str) the screen name
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        The json result of the method call
    """
    try:
        endpoint = f"/openc3-api/screen/{target_name.upper()}/{screen_name.upper()}"
        response = openc3.script.API_SERVER.request("delete", endpoint, scope=scope)
        if not response or response.status_code != 200:
            if response:
                parsed = json.loads(response.text)
                raise RuntimeError(f"delete_screen error: {parsed['error']}")
            else:
                raise RuntimeError("delete_screen failed")

        return response.text
    except Exception as error:
        raise RuntimeError(f"delete_screen failed due to {repr(error)}") from error


def display_screen(target_name, screen_name, x=None, y=None, scope=OPENC3_SCOPE):
    # Noop outside of ScriptRunner
    pass


def clear_screen(target_name, screen_name):
    # Noop outside of ScriptRunner
    pass


def clear_all_screens():
    # Noop outside of ScriptRunner
    pass


def local_screen(screen_name, definition, x=None, y=None):
    # Noop outside of ScriptRunner
    pass



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.config.config_parser import ConfigParser
from openc3.interfaces.protocols.burst_protocol import BurstProtocol
from openc3.utilities.extract import hex_to_byte_string


# Protocol which delineates packets using termination characters at
# the end of the data.
class TerminatedProtocol(BurstProtocol):
    # self.param write_termination_characters [String] The characters to write
    #   after writing the Packet buffer. Must be given as a
    #   hexadecimal string such as '0xABCD'.
    # self.param read_termination_characters [String] The characters at the end of
    #   the data which delineate the end of a Packet. Must be given as a
    #   hexadecimal string such as '0xABCD'.
    # self.param strip_read_termination [Boolean] Whether to remove the
    #   read_termination_characters before turning the data into a
    #   Packet.
    # self.param discard_leading_bytes (see BurstProtocol#initialize)
    # self.param sync_pattern (see BurstProtocol#initialize)
    # self.param fill_fields (see BurstProtocol#initialize)
    # self.param allow_empty_data [True/False/None] See Protocol#initialize
    def __init__(
        self,
        write_termination_characters,
        read_termination_characters,
        strip_read_termination=True,
        discard_leading_bytes=0,
        sync_pattern=None,
        fill_fields=False,
        allow_empty_data=None,
    ):
        self.write_termination_characters = hex_to_byte_string(write_termination_characters)
        self.read_termination_characters = hex_to_byte_string(read_termination_characters)
        self.strip_read_termination = ConfigParser.handle_true_false(strip_read_termination)
        if self.strip_read_termination is not True and self.strip_read_termination is not False:
            raise RuntimeError("strip_read_termination must be True or False")

        super().__init__(discard_leading_bytes, sync_pattern, fill_fields, allow_empty_data)

    def write_data(self, data, extra):
        try:
            data.index(self.write_termination_characters)
            raise RuntimeError("Packet contains termination characters!")
        except ValueError:  # no index
            pass

        data, extra = super().write_data(data, extra)
        for byte in self.write_termination_characters:
            data.append(byte)
        return (data, extra)

    def reduce_to_single_packet(self):
        try:
            index = self.data.index(self.read_termination_characters)

            # Reduce to packet data and setup current_data for next packet
            if index > 0:
                if self.strip_read_termination:
                    packet_data = self.data[0:index]
                else:
                    packet_data = self.data[0 : (index + len(self.read_termination_characters))]
            else:  # self.data begins with the termination characters
                if self.strip_read_termination:
                    packet_data = b""
                else:  # Keep everything
                    packet_data = self.data[0 : (len(self.read_termination_characters))]
            self.data = self.data[(index + len(self.read_termination_characters)) :]
            return (packet_data, self.extra)
        except ValueError:  # sync_index = None
            return ("STOP", self.extra)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import queue
import requests
from openc3.config.config_parser import ConfigParser
from openc3.interfaces.interface import Interface
from openc3.accessors.http_accessor import HttpAccessor
from openc3.packets.packet import Packet


class HttpClientInterface(Interface):
    """
    HttpClientInterface is a class that provides an interface for making HTTP
    requests using the requests library.
    """

    def __init__(
        self,
        hostname,
        port=80,
        protocol="http",
        write_timeout=None,
        read_timeout=None,
        connect_timeout=5,
        include_request_in_response=False,
    ):
        """
        Initializes the HTTPClientInterface with the given parameters.
        Args:
            hostname (str): The hostname of the server.
            port (int, optional): The port number to connect to. Defaults to 80.
            protocol (str, optional): The protocol to use ('http' or 'https'). Defaults to "http".
            write_timeout (None, optional): Present to match Ruby parameters but not used.
            read_timeout (float or None, optional): The timeout for reading operations in seconds. Defaults to None.
            connect_timeout (float or None, optional): The timeout for connection operations in seconds. Defaults to 5.
            include_request_in_response (bool, optional): Whether to include the request in the response. Defaults to False.
        """
        super().__init__()
        self.hostname = hostname
        self.port = int(port)
        self.protocol = protocol
        if (self.port == 80 and self.protocol == "http") or (self.port == 443 and self.protocol == "https"):
            self.url = f"{self.protocol}://{self.hostname}"
        else:
            self.url = f"{self.protocol}://{self.hostname}:{self.port}"
        self.read_timeout = ConfigParser.handle_none(read_timeout)
        if self.read_timeout:
            self.read_timeout = float(self.read_timeout)
        self.connect_timeout = ConfigParser.handle_none(connect_timeout)
        if self.connect_timeout:
            self.connect_timeout = float(self.connect_timeout)
        self.include_request_in_response = ConfigParser.handle_true_false(include_request_in_response)

        self.response_queue = queue.Queue()

    def connection_string(self):
        """Returns the url."""
        return self.url

    def connect(self):
        """
        Initializes an HTTP session and then calls the parent class's connect method.
        """
        self.http = requests.Session()
        super().connect()

    def connected(self):
        """
        Check if the HTTP client is connected.

        Returns:
            bool: True if the HTTP client is connected, False otherwise.
        """
        if self.http:
            return True
        else:
            return False

    # Disconnects the interface from its target(s)
    def disconnect(self):
        """
        Disconnects the HTTP client interface.

        This method closes the HTTP connection if it exists, sets the HTTP client to None,
        clears the response queue, calls the superclass's disconnect method, and unblocks
        the response queue to allow the read_interface method to return.
        """
        if self.http:
            self.http.close
        self.http = None
        while not self.response_queue.empty():
            self.response_queue.get_nowait()
        super().disconnect()
        self.response_queue.put((None, None))

    def convert_packet_to_data(self, packet):
        """
        Converts a packet to data and extracts additional information.

        Args:
            packet (Packet): The packet to be converted.
        Returns:
            tuple: A tuple containing:
                - data (bytes): The buffer data from the packet.
                - extra (dict): A dictionary containing additional information extracted from the packet.
                    - "HTTP_URI" (str): The full HTTP URI constructed from the base URL and the packet's HTTP path.
                    - "HTTP_REQUEST_TARGET_NAME" (str): The target name of the HTTP request.
        """

        extra = packet.extra
        extra = extra or {}
        data = packet.buffer  # Copy buffer so logged command isn't modified
        extra["HTTP_URI"] = f"{self.url}{packet.read('HTTP_PATH')}"
        extra["HTTP_REQUEST_TARGET_NAME"] = packet.target_name
        return data, extra

    # Calls the http request method to send the data to the target
    def write_interface(self, data, extra=None):
        """
        Sends the data to the target using an HTTP request.

        Args:
            data (bytes): The data to be sent.
            extra (dict, optional): Additional parameters for the HTTP request. Defaults to None.
                - HTTP_QUERIES: Query parameters for the HTTP request.
                - HTTP_HEADERS: Headers for the HTTP request.
                - HTTP_URI: The URI for the HTTP request.
                - HTTP_METHOD: The HTTP method (e.g., GET, POST).

        Returns:
            tuple: The data and extra parameters.
        """
        extra = extra or {}
        params = extra.get("HTTP_QUERIES")
        headers = extra.get("HTTP_HEADERS")
        uri = extra["HTTP_URI"]
        method = extra["HTTP_METHOD"]

        resp = self.http.request(
            method,
            uri,
            params=params,
            headers=headers,
            data=data,
            stream=False,
            timeout=(self.connect_timeout, self.read_timeout),
        )

        # Normalize Response into simple hash
        response_data = None
        response_extra = {}
        response_extra["HTTP_REQUEST"] = [data, extra]
        if resp.headers and len(resp.headers) > 0:
            # Cast headers to a dictionary so it can be serialized
            # because the request library returns CaseInsensitiveDict
            response_extra["HTTP_HEADERS"] = dict(resp.headers)
        response_extra["HTTP_STATUS"] = resp.status_code
        response_data = bytearray(resp.text, encoding="utf-8")

        self.response_queue.put((response_data, response_extra))
        self.write_interface_base(data, extra)
        return data, extra

    def read_interface(self):
        """
        Returns the response data and extra parameters from the interface,
        which were queued up by the write_interface method.
        Read protocols can then potentially modify the data in their read_data methods.
        Then convert_data_to_packet is called to convert the data into a Packet object.
        Finally the read protocols read_packet methods are called.

        Returns:
            tuple: The response data and extra parameters.
        """
        data, extra = self.response_queue.get(block=True)
        if data is None:
            return data, extra
        self.read_interface_base(data, extra)
        return data, extra

    def convert_data_to_packet(self, data, extra=None):
        """
        Converts the read data into an OpenC3 Packet object.

        Args:
            data (str): Raw packet data.
            extra (dict, optional): Additional parameters for the packet. Defaults to None.
                - HTTP_HEADERS: Hash of response headers.
                - HTTP_STATUS: Integer response status code.
                - HTTP_REQUEST: [data, extra] where data is the request data and extra contains:
                    - HTTP_REQUEST_TARGET_NAME: String request target name.
                    - HTTP_URI: String request URI based on HTTP_PATH.
                    - HTTP_PATH: String request path.
                    - HTTP_METHOD: String request method.
                    - HTTP_PACKET: String response packet name.
                    - HTTP_ERROR_PACKET: Optional string error packet name.
                    - HTTP_QUERIES: Optional hash of request queries.
                    - HTTP_HEADERS: Optional hash of request headers.

        Returns:
            Packet: OpenC3 Packet with buffer filled with data.
        """
        packet = Packet(None, None, "BIG_ENDIAN", None, data)
        packet.accessor = HttpAccessor(packet)
        # Grab the request extra set in the write_interface method
        request_extra = None
        if extra and extra["HTTP_REQUEST"]:
            request_extra = extra["HTTP_REQUEST"][1]
        if request_extra is not None:
            # Identify the response
            request_target_name = request_extra.get("HTTP_REQUEST_TARGET_NAME")
            if request_target_name is not None:
                request_target_name = str(request_target_name).upper()
                response_packet_name = request_extra.get("HTTP_PACKET")
                error_packet_name = request_extra.get("HTTP_ERROR_PACKET")
                # HTTP_STATUS was set in the base extra
                status = int(extra["HTTP_STATUS"])
                if status >= 300 and error_packet_name is not None:
                    # Handle error special case response packet
                    packet.target_name = request_target_name
                    packet.packet_name = str(error_packet_name).upper()
                elif response_packet_name is not None:
                    packet.target_name = request_target_name
                    packet.packet_name = str(response_packet_name).upper()

            if not self.include_request_in_response:
                extra.pop("HTTP_REQUEST", None)
        packet.extra = extra
        return packet



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.environment import OPENC3_LOCAL_MODE, OPENC3_CONFIG_BUCKET
from openc3.utilities.local_mode import LocalMode
from openc3.utilities.bucket import Bucket


class TargetFile:
    @classmethod
    def body(cls, scope, name):
        name = name.split("*")[0]  # Split '*' that indicates modified
        # First try opening a potentially modified version by looking for the modified target
        if OPENC3_LOCAL_MODE:
            local_file = LocalMode.open_local_file(name, scope=scope)
            if local_file:
                return local_file.read()

        bucket = Bucket.getClient()
        resp = bucket.get_object(bucket=OPENC3_CONFIG_BUCKET, key=f"{scope}/targets_modified/{name}")
        if not resp:
            # Now try the original
            resp = bucket.get_object(bucket=OPENC3_CONFIG_BUCKET, key=f"{scope}/targets/{name}")
        if resp and resp["Body"]:
            return resp["Body"].read()
        else:
            return None



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.environment import OPENC3_SCOPE
from openc3.models.model import Model


class SecretModel(Model):
    PRIMARY_KEY = "openc3__secrets"

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name: str, scope: str):
        return super().get(f"{scope}__{SecretModel.PRIMARY_KEY}", name)

    @classmethod
    def names(cls, scope: str):
        return super().names(f"{scope}__{SecretModel.PRIMARY_KEY}")

    @classmethod
    def all(cls, scope: str):
        return super().all(f"{scope}__{SecretModel.PRIMARY_KEY}")

    # END NOTE

    def __init__(self, name: str, value: str, scope: str = OPENC3_SCOPE):
        super().__init__(f"{scope}__{SecretModel.PRIMARY_KEY}", name=name, scope=scope)
        self.value: str = value

    # @return [Hash] JSON encoding of this model
    def as_json(self):
        return {
            "name": self.name,
            "value": self.value,
        }



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


class Limits:
    """Limits uses PacketConfig to parse the command and telemetry
    configuration files. It provides the API layer which other classes use to
    access information about and manipulate limits. This includes getting,
    setting and checking individual limit items as well as manipulating limits groups.
    """

    LATEST_PACKET_NAME = "LATEST"

    # @param config [PacketConfig] Packet configuration to use to access the
    #  limits
    def __init__(self, config, system):
        self.config = config
        self.system = system

    # (see PacketConfig#warnings)
    def warnings(self):
        return self.config.warnings

    # (see PacketConfig#limits_sets)
    def sets(self):
        return self.config.limits_sets

    # @return [Hash(String, Array)] The defined limits groups
    def groups(self):
        return self.config.limits_groups

    # Checks whether the limits are enabled for the specified item
    #
    # @param target_name [String] The target name
    # @param packet_name [String] The packet name. Must be a defined packet name and not 'LATEST'.
    # @param item_name [String] The item name
    def enabled(self, target_name, packet_name, item_name):
        return self._get_packet(target_name, packet_name).get_item(item_name).limits.enabled

    # Enables limit checking for the specified item
    #
    # @param (see #enabled?)
    def enable(self, target_name, packet_name, item_name):
        self._get_packet(target_name, packet_name).enable_limits(item_name)

    # Disables limit checking for the specified item
    #
    # @param (see #enabled?)
    def disable(self, target_name, packet_name, item_name):
        self._get_packet(target_name, packet_name).disable_limits(item_name)

    # Get the limits for a telemetry item
    #
    # @param target_name [String] Target Name
    # @param packet_name [String] Packet Name
    # @param item_name [String] Item Name
    # @param limits_set [String or Symbol or nil] Desired Limits set.  nil = current limits set
    # @return [Array<limits_set, persistence, enabled, red_low, yellow_low, red_high, yellow_high, green_low (optional), green_high (optional)] Limits information
    def get(self, target_name, packet_name, item_name, limits_set=None):
        limits = self._get_packet(target_name, packet_name).get_item(item_name).limits
        if limits.values:
            if limits_set:
                limits_set = str(limits_set).upper()
            else:
                limits_set = self.system.limits_set()
            limits_for_set = limits.values.get(limits_set)
            if limits_for_set is not None:
                gl = None
                gh = None
                if len(limits_for_set) > 4:
                    gl = limits_for_set[4]
                    gh = limits_for_set[5]
                return [
                    limits_set,
                    limits.persistence_setting,
                    limits.enabled,
                    limits_for_set[0],
                    limits_for_set[1],
                    limits_for_set[2],
                    limits_for_set[3],
                    gl, gh,
                ]
            else:
                return [None, None, None, None, None, None, None, None, None]
        else:
            return [None, None, None, None, None, None, None, None, None]

    # Set the limits for a telemetry item
    #
    # @param target_name [String] Target Name
    # @param packet_name [String] Packet Name
    # @param item_name [String] Item Name
    # @param red_low [Float] Red Low Limit
    # @param yellow_low [Float] Yellow Low Limit
    # @param yellow_high [Float] Yellow High Limit
    # @param red_high [Float] Red High Limit
    # @param green_low [Float] Green Low Limit
    # @param green_high [Float] Green High Limit
    # @param limits_set [String or Symbol or nil] Desired Limits set.  nil = current limits set, recommend using :CUSTOM
    # @param persistence [Integer] The number of samples the value must be out of limits before detecting a limits change. nil = Leave unchanged
    # @param enabled [Boolean] If limits monitoring is enabled for this item
    # @return [Array<limits_set, persistence, enabled, red_low, yellow_low, red_high, yellow_high, green_low (optional), green_high (optional)] Limits information
    def set(
        self,
        target_name,
        packet_name,
        item_name,
        red_low,
        yellow_low,
        yellow_high,
        red_high,
        green_low=None,
        green_high=None,
        limits_set="CUSTOM",
        persistence=None,
        enabled=True,
    ):
        packet = self._get_packet(target_name, packet_name)
        item = packet.get_item(item_name)
        limits = item.limits
        if limits_set:
            limits_set = str(limits_set).upper()
        else:
            limits_set = self.system.limits_set()
        if limits.values is None:
            if limits_set == "DEFAULT":
                limits.values = {"DEFAULT": []}
            else:
                raise RuntimeError(
                    f"DEFAULT limits must be defined for {target_name} {packet_name} {item_name} before setting limits set {limits_set}"
                )
        limits_for_set = limits.values.get(limits_set)
        if limits_for_set is None:
            limits.values[limits_set] = []
            limits_for_set = limits.values[limits_set]
        while len(limits_for_set) < 6:
            limits_for_set.append(None)
        limits_for_set[0] = float(red_low)
        limits_for_set[1] = float(yellow_low)
        limits_for_set[2] = float(yellow_high)
        limits_for_set[3] = float(red_high)
        if green_low and green_high:
            limits_for_set[4] = float(green_low)
            limits_for_set[5] = float(green_high)
        else:
            limits_for_set[4] = None
            limits_for_set[5] = None
        if enabled is not None:
            limits.enabled = enabled
        if persistence is not None:
            limits.persistence_setting = int(persistence)
        packet.update_limits_items_cache(item)
        if limits_set not in self.config.limits_sets:
            self.config.limits_sets.append(limits_set)
        return [
            limits_set,
            limits.persistence_setting,
            limits.enabled,
            limits_for_set[0],
            limits_for_set[1],
            limits_for_set[2],
            limits_for_set[3],
            limits_for_set[4],
            limits_for_set[5],
        ]

    def _get_packet(self, target_name, packet_name):
        if packet_name.upper() == Limits.LATEST_PACKET_NAME:
            raise RuntimeError("LATEST packet not valid")

        packets = self.config.telemetry.get(target_name.upper())
        if packets is None:
            raise RuntimeError(f"Telemetry target '{target_name.upper()}' does not exist")

        packet = packets.get(packet_name.upper())
        if packet is None:
            raise RuntimeError(f"Telemetry packet '{target_name.upper()} { packet_name.upper()}' does not exist")

        return packet



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.
import io
import os
import json
import tempfile
import requests
from openc3.utilities.extract import *
import openc3.script
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.local_mode import LocalMode

OPENC3_CLOUD = os.environ.get("OPENC3_CLOUD") or "local"


def delete_target_file(path: str, scope: str = OPENC3_SCOPE):
    """Delete a file on a target

    Args:
        path (str) Path to a file in a target directory
        scope (str) Optional, defaults to env.OPENC3_SCOPE
    """
    try:
        # Only delete from the targets_modified
        delete_path = f"{scope}/targets_modified/{path}"
        endpoint = f"/openc3-api/storage/delete/{delete_path}"
        print(f"Deleting {delete_path}")
        # Pass the name of the ENV variable name where we pull the actual bucket name
        response = openc3.script.API_SERVER.request(
            "delete", endpoint, query={"bucket": "OPENC3_CONFIG_BUCKET"}, scope=scope
        )
        if not response or response.status_code != 200:
            raise Exception(f"Failed to delete {delete_path}")
    except Exception as error:
        raise Exception(f"Failed deleting {path} due to {repr(error)}") from error
    return None


def put_target_file(path: str, io_or_string: io.IOBase | str, scope: str = OPENC3_SCOPE):
    """Get a handle to write a target file

    Args:
        path (str) Path to a file in a target directory
        io_or_string (io | str) IO object or str object
        scope (str) Optional, defaults to env.OPENC3_SCOPE
    """
    if ".." in path:
        raise Exception(f"Disallowed path modifier '..' found in {path}")

    upload_path = f"{scope}/targets_modified/{path}"

    if os.getenv("OPENC3_LOCAL_MODE") and openc3.script.OPENC3_IN_CLUSTER:
        LocalMode.put_target_file(upload_path, io_or_string, scope=scope)
        if hasattr(io_or_string, "read"):  # not str or bytes
            io_or_string.seek(0)

    endpoint = f"/openc3-api/storage/upload/{upload_path}"
    result = _get_presigned_request(endpoint, scope=scope)
    print(f"Writing {upload_path}")

    # Try to put the file
    try:
        uri = _get_uri(result["url"])
        with requests.Session() as s:
            if hasattr(io_or_string, "read"):
                # TODO: Better way to get io size?
                io_or_string.seek(0, 2)  # Jump to end
                length = io_or_string.tell()
                io_or_string.seek(0)
                io_or_string = io_or_string.read()
            else:  # str or bytes
                length = len(io_or_string)
            result = s.put(
                uri,
                data=io_or_string,
                headers={"Content-Length": str(length)},
            )
            return result.content
    except Exception as error:
        raise Exception(f"Failed to write {upload_path} due to {repr(error)}") from error


def get_target_file(path: str, original: bool = False, scope: str = OPENC3_SCOPE):
    """Get a handle to access a target file

    Args:
        path (str) Path to a file in a target directory, e.g. "INST/procedures/test.rb"
        original (bool) Whether to get the original or modified file; defaults to False
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        (File | None)
    """
    part = "targets"
    if original is False:
        part += "_modified"
    # Loop to allow redo when switching from modified to original
    while True:
        try:
            if part == "targets_modified" and os.getenv("OPENC3_LOCAL_MODE"):
                local_file = LocalMode.open_local_file(path, scope=scope)
                if local_file:
                    print(f"Reading local {scope}/{part}/{path}")
                    file = tempfile.NamedTemporaryFile(mode="w+b")
                    file.write(local_file.read())
                    file.seek(0)  # Rewind so the file is ready to read
                    return file
            return _get_storage_file(f"{part}/{path}", scope=scope)
        except Exception as error:
            if part == "targets_modified":
                part = "targets"
                # redo
            else:
                raise error


# These are helper methods ... should not be used directly


def _get_download_url(path: str, scope: str = OPENC3_SCOPE):
    """Get a download url for object in block storage

    Args:
        path (str) Path to a file in a target directory, e.g. "INST/procedures/test.rb"
        scope (str) Optional, defaults to env.OPENC3_SCOPE
    """
    targets = "targets_modified"  # First try targets_modified
    response = openc3.script.API_SERVER.request(
        "get",
        f"/openc3-api/storage/exists/{scope}/{targets}/{path}",
        query={"bucket": "OPENC3_CONFIG_BUCKET"},
        scope=scope,
    )

    if response.status_code != 200:
        targets = "targets"  # Next try targets
        response = openc3.script.API_SERVER.request(
            "get",
            f"/openc3-api/storage/exists/{scope}/{targets}/{path}",
            query={"bucket": "OPENC3_CONFIG_BUCKET"},
            scope=scope,
        )
        if response.status_code != 200:
            raise RuntimeError(f"File not found: {path} in scope: {scope}")

    endpoint = f"/openc3-api/storage/download/{scope}/{targets}/{path}"
    # external must be true because we're using this URL from the frontend
    result = _get_presigned_request(endpoint, external=True, scope=scope)
    return result["url"]


def _get_storage_file(path, scope=OPENC3_SCOPE):
    # Create Tempfile to store data
    file = tempfile.NamedTemporaryFile(mode="w+b")

    endpoint = f"/openc3-api/storage/download/{scope}/{path}"
    result = _get_presigned_request(endpoint, scope=scope)
    print(f"Reading {scope}/{path}")

    # Try to get the file
    uri = _get_uri(result["url"])
    response = requests.get(uri)
    if response.status_code == 404:
        raise RuntimeError(f"File not found: {scope}/{path}")
    file.write(response.content)
    file.seek(0)
    return file


def _get_uri(url):
    if openc3.script.OPENC3_IN_CLUSTER:
        match OPENC3_CLOUD:
            case "local":
                bucket_url = os.environ.get("OPENC3_BUCKET_URL", "http://openc3-minio:9000")
                return f"{bucket_url}{url}"
            case "aws":
                return f"https://s3.{os.getenv('AWS_REGION')}.amazonaws.com{url}"
            case "gcp":
                return f"https://storage.googleapis.com{url}"
            # when 'azure'
            case _:
                raise Exception(f"Unknown cloud {OPENC3_CLOUD}")
    else:
        return f"{openc3.script.API_SERVER.generate_url()}{url}"


def _get_presigned_request(endpoint, external=None, scope=OPENC3_SCOPE):
    if external or not openc3.script.OPENC3_IN_CLUSTER:
        response = openc3.script.API_SERVER.request(
            "get", endpoint, query={"bucket": "OPENC3_CONFIG_BUCKET"}, scope=scope
        )
    else:
        response = openc3.script.API_SERVER.request(
            "get",
            endpoint,
            query={"bucket": "OPENC3_CONFIG_BUCKET", "internal": True},
            scope=scope,
        )

    if not response or response.status_code != 201:
        raise Exception(f"Failed to get presigned URL for {endpoint}")
    return json.loads(response.text)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import time
from openc3.interfaces.interface import Interface
from openc3.utilities.logger import Logger
from openc3.top_level import get_class_from_module
from openc3.utilities.string import filename_to_module, filename_to_class_name


# An interface class that provides simulated telemetry and command responses
class SimulatedTargetInterface(Interface):
    # @param sim_target_file [String] Filename of the simulator target class
    def __init__(self, sim_target_file):
        super().__init__()
        self.__connected = False
        self.initialized = False
        self.count_100hz = 0
        self.next_tick_time = None
        self.pending_packets = []
        self.sim_target_class = get_class_from_module(
            filename_to_module(sim_target_file),
            filename_to_class_name(sim_target_file),
        )
        self.sim_target = None
        self.write_raw_allowed = False

    def connection_string(self):
        return self.sim_target_class.__name__

    # Initialize the simulated target object and "connect" to the target
    def connect(self):
        if not self.initialized:
            # Create Simulated Target Object
            self.sim_target = self.sim_target_class(self.target_names[0])
            # Set telemetry rates
            self.sim_target.set_rates()

            self.initialized = True

        self.count_100hz = 0

        # Save the current time + delta as the next expected tick time
        self.next_tick_time = time.time() + self.sim_target.tick_period_seconds()

        super().connect()
        self.__connected = True

    # @return [Boolean] Whether the simulated target is connected (initialized)
    def connected(self):
        return self.__connected

    # @return [Packet] Returns a simulated target packet from the simulator
    def read(self):
        packet = None
        if self.connected():
            while True:
                packet = self.first_pending_packet()
                if packet is None:
                    break
                # Support read_packet (but not read data) in protocols
                # Generic protocol use is not supported
                for protocol in self.read_protocols:
                    packet = protocol.read_packet(packet)
                    if packet == "DISCONNECT":
                        Logger.info(
                            f"{self.name}: Protocol {protocol.__class__.__name__} read_packet requested disconnect"
                        )
                        return None
                    if packet == "STOP":
                        break
                if packet != "STOP":
                    return packet

            while True:
                # Calculate time to sleep to make ticks the right distance apart
                now = time.time()
                delta = self.next_tick_time - now
                if delta > 0.0:
                    time.sleep(delta)  # Sleep between packets
                    if not self.connected():
                        return None
                elif delta < -1.0:
                    # Fell way behind - jump next tick time
                    self.next_tick_time = time.time()

                self.pending_packets = self.sim_target.read(self.count_100hz, self.next_tick_time)
                self.next_tick_time += self.sim_target.tick_period_seconds()
                self.count_100hz += self.sim_target.tick_increment()

                packet = self.first_pending_packet()
                if packet:
                    # Support read_packet (but not read data) in protocols
                    # Generic protocol use is not supported
                    for protocol in self.read_protocols:
                        packet = protocol.read_packet(packet)
                        if packet == "DISCONNECT":
                            Logger.info(
                                f"{self.name}: Protocol {protocol.__class__.__name__} read_packet requested disconnect"
                            )
                            return None
                        if packet == "STOP":
                            break
                    if packet == "STOP":
                        continue
                    return packet

        else:
            raise RuntimeError("Interface not connected")

    # @param packet [Packet] Command packet to send to the simulator
    def write(self, packet):
        if self.connected():
            # Update count of commands sent through this interface
            self.write_count += 1
            self.bytes_written += len(packet.buffer)
            self.written_raw_data_time = time.time()
            self.written_raw_data = packet.buffer

            # Have simulated target handle the packet
            self.sim_target.write(packet)
        else:
            raise RuntimeError("Interface not connected")

    # write_raw is not implemented and will raise a RuntimeError
    def write_raw(self, data):
        raise RuntimeError("write_raw not implemented for SimulatedTargetInterface")

    # Disconnect from the simulator
    def disconnect(self):
        self.__connected = False
        super().disconnect()

    def first_pending_packet(self):
        packet = None
        if len(self.pending_packets) != 0:
            self.read_count += 1
            packet = self.pending_packets.pop(0).clone()
            self.bytes_read += len(packet.buffer)
            self.read_raw_data_time = time.time()
            self.read_raw_data = packet.buffer
        return packet



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


from openc3.conversions.conversion import Conversion
from openc3.utilities.time import formatted


# Converts the packet received time object into a formatted string.
class ReceivedTimeFormattedConversion(Conversion):
    # Initializes converted_type to 'STRING' and converted_bit_size to 0
    def __init__(self):
        super().__init__()
        self.converted_type = "STRING"
        self.converted_bit_size = 0

    # self.param (see Conversion#call)
    # self.return [String] Formatted packet time
    def call(self, value, packet, buffer):
        if packet.received_time is not None:
            return formatted(packet.received_time)
        else:
            return "No Packet Received Time"
